import asyncio
from collections.abc import AsyncIterable
from typing import Union

from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow._constants import ERR_UDF_EXCEPTION_STRING
from pynumaflow.proto.reducer import reduce_pb2, reduce_pb2_grpc
from pynumaflow.reducestreamer._dtypes import (
    Datum,
    ReduceStreamAsyncCallable,
    _ReduceStreamBuilderClass,
    ReduceRequest,
)
from pynumaflow.reducestreamer.servicer.task_manager import TaskManager
from pynumaflow.shared.server import handle_async_error
from pynumaflow.types import NumaflowServicerContext


async def datum_generator(
    request_iterator: AsyncIterable[reduce_pb2.ReduceRequest],
) -> AsyncIterable[ReduceRequest]:
    """Generate a ReduceRequest from a ReduceRequest proto message."""
    async for d in request_iterator:
        reduce_request = ReduceRequest(
            operation=d.operation.event,
            windows=d.operation.windows,
            payload=Datum(
                keys=list(d.payload.keys),
                value=d.payload.value,
                event_time=d.payload.event_time.ToDatetime(),
                watermark=d.payload.watermark.ToDatetime(),
                headers=dict(d.payload.headers),
            ),
        )
        yield reduce_request


class AsyncReduceStreamServicer(reduce_pb2_grpc.ReduceServicer):
    """
    This class is used to create a new grpc Reduce servicer instance.
    Provides the functionality for the required rpc methods.
    """

    def __init__(
        self,
        handler: Union[ReduceStreamAsyncCallable, _ReduceStreamBuilderClass],
    ):
        # The Reduce handler can be a function or a builder class instance.
        self.__reduce_handler: Union[ReduceStreamAsyncCallable, _ReduceStreamBuilderClass] = handler

    async def ReduceFn(
        self,
        request_iterator: AsyncIterable[reduce_pb2.ReduceRequest],
        context: NumaflowServicerContext,
    ) -> reduce_pb2.ReduceResponse:
        """
        Applies a reduce function to a datum stream.
        The pascal case function name comes from the proto reduce_pb2_grpc.py file.
        """
        # Create a task manager instance
        task_manager = TaskManager(handler=self.__reduce_handler)

        # Create a consumer task to read from the result queue
        # All the results from the reduce function will be sent to the result queue
        # We will read from the result queue and send the results to the client
        consumer = task_manager.global_result_queue.read_iterator()

        # Create an async iterator from the request iterator
        datum_iterator = datum_generator(request_iterator=request_iterator)

        # Create a process_input_stream task in the task manager,
        # this would read from the datum iterator
        # and then create the required tasks to process the data requests
        # The results from these tasks are then sent to the result queue
        producer = asyncio.create_task(task_manager.process_input_stream(datum_iterator))

        # Start the consumer task where we read from the result queue
        # and send the results to the client
        # The task manager can write the following to the result queue:
        # 1. A reduce_pb2.ReduceResponse message
        # This is the result of the reduce function, it contains the window and the
        # result of the reduce function
        # The result of the reduce function is a reduce_pb2.ReduceResponse message and can be
        # directly sent to the client
        #
        # 2. An Exception
        # Any exceptions that occur during the processing reduce function tasks are
        # sent to the result queue. We then forward these exception to the client
        #
        # 3. A reduce_pb2.ReduceResponse message with EOF=True
        # This is a special message that indicates the end of the processing for a window
        # When we get this message, we send an EOF message to the client
        try:
            async for msg in consumer:
                # If the message is an exception, we raise the exception
                if isinstance(msg, BaseException):
                    await handle_async_error(context, msg, ERR_UDF_EXCEPTION_STRING)
                    return
                # Send window EOF response or Window result response
                # back to the client
                else:
                    yield msg
        except BaseException as e:
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)
            return
        # Wait for the process_input_stream task to finish for a clean exit
        try:
            await producer
        except BaseException as e:
            await handle_async_error(context, e, ERR_UDF_EXCEPTION_STRING)
            return

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> reduce_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto reduce_pb2_grpc.py file.
        """
        return reduce_pb2.ReadyResponse(ready=True)



from pynumaflow.reducestreamer._dtypes import (
    Message,
    Datum,
    IntervalWindow,
    Metadata,
    DROP,
    ReduceStreamer,
    ReduceWindow,
)
from pynumaflow.reducestreamer.async_server import ReduceStreamAsyncServer

__all__ = [
    "Message",
    "Datum",
    "IntervalWindow",
    "Metadata",
    "DROP",
    "ReduceStreamAsyncServer",
    "ReduceStreamer",
    "ReduceWindow",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/sourcetransformer/transform.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/sourcetransformer/transform.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2pynumaflow/proto/sourcetransformer/transform.proto\x12\x14sourcetransformer.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto\"\x18\n\tHandshake\x12\x0b\n\x03sot\x18\x01 \x01(\x08\"\xbe\x03\n\x16SourceTransformRequest\x12\x45\n\x07request\x18\x01 \x01(\x0b\x32\x34.sourcetransformer.v1.SourceTransformRequest.Request\x12\x37\n\thandshake\x18\x02 \x01(\x0b\x32\x1f.sourcetransformer.v1.HandshakeH\x00\x88\x01\x01\x1a\x95\x02\n\x07Request\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12R\n\x07headers\x18\x05 \x03(\x0b\x32\x41.sourcetransformer.v1.SourceTransformRequest.Request.HeadersEntry\x12\n\n\x02id\x18\x06 \x01(\t\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0c\n\n_handshake\"\x98\x02\n\x17SourceTransformResponse\x12\x45\n\x07results\x18\x01 \x03(\x0b\x32\x34.sourcetransformer.v1.SourceTransformResponse.Result\x12\n\n\x02id\x18\x02 \x01(\t\x12\x37\n\thandshake\x18\x03 \x01(\x0b\x32\x1f.sourcetransformer.v1.HandshakeH\x00\x88\x01\x01\x1a\x63\n\x06Result\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04tags\x18\x04 \x03(\tB\x0c\n\n_handshake\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\xcf\x01\n\x0fSourceTransform\x12t\n\x11SourceTransformFn\x12,.sourcetransformer.v1.SourceTransformRequest\x1a-.sourcetransformer.v1.SourceTransformResponse(\x01\x30\x01\x12\x46\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a#.sourcetransformer.v1.ReadyResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.sourcetransformer.transform_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SOURCETRANSFORMREQUEST_REQUEST_HEADERSENTRY']._loaded_options = None
  _globals['_SOURCETRANSFORMREQUEST_REQUEST_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_HANDSHAKE']._serialized_start=138
  _globals['_HANDSHAKE']._serialized_end=162
  _globals['_SOURCETRANSFORMREQUEST']._serialized_start=165
  _globals['_SOURCETRANSFORMREQUEST']._serialized_end=611
  _globals['_SOURCETRANSFORMREQUEST_REQUEST']._serialized_start=320
  _globals['_SOURCETRANSFORMREQUEST_REQUEST']._serialized_end=597
  _globals['_SOURCETRANSFORMREQUEST_REQUEST_HEADERSENTRY']._serialized_start=551
  _globals['_SOURCETRANSFORMREQUEST_REQUEST_HEADERSENTRY']._serialized_end=597
  _globals['_SOURCETRANSFORMRESPONSE']._serialized_start=614
  _globals['_SOURCETRANSFORMRESPONSE']._serialized_end=894
  _globals['_SOURCETRANSFORMRESPONSE_RESULT']._serialized_start=781
  _globals['_SOURCETRANSFORMRESPONSE_RESULT']._serialized_end=880
  _globals['_READYRESPONSE']._serialized_start=896
  _globals['_READYRESPONSE']._serialized_end=926
  _globals['_SOURCETRANSFORM']._serialized_start=929
  _globals['_SOURCETRANSFORM']._serialized_end=1136
# @@protoc_insertion_point(module_scope)



from pynumaflow.sourcer._dtypes import (
    Message,
    ReadRequest,
    PendingResponse,
    AckRequest,
    NackRequest,
    Offset,
    PartitionsResponse,
    get_default_partitions,
    Sourcer,
    SourceCallable,
)
from pynumaflow.sourcer.async_server import SourceAsyncServer

__all__ = [
    "Message",
    "ReadRequest",
    "PendingResponse",
    "AckRequest",
    "NackRequest",
    "Offset",
    "PartitionsResponse",
    "get_default_partitions",
    "Sourcer",
    "SourceAsyncServer",
    "SourceCallable",
]



import asyncio
from collections.abc import AsyncIterator

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from pynumaflow.shared.asynciter import NonBlockingIterator

from pynumaflow.shared.server import handle_async_error
from pynumaflow.sourcer import ReadRequest, Offset, NackRequest, AckRequest, SourceCallable
from pynumaflow.proto.sourcer import source_pb2
from pynumaflow.proto.sourcer import source_pb2_grpc
from pynumaflow.types import NumaflowServicerContext
from pynumaflow._constants import _LOGGER, STREAM_EOF, ERR_UDF_EXCEPTION_STRING


def _create_read_handshake_response():
    """Create a handshake response for the Read function."""
    return source_pb2.ReadResponse(
        status=source_pb2.ReadResponse.Status(
            eot=False, code=source_pb2.ReadResponse.Status.SUCCESS
        ),
        handshake=source_pb2.Handshake(sot=True),
    )


def _create_ack_handshake_response():
    """Create a handshake response for the Ack function."""
    return source_pb2.AckResponse(
        result=source_pb2.AckResponse.Result(success=_empty_pb2.Empty()),
        handshake=source_pb2.Handshake(sot=True),
    )


def _create_read_response(response):
    """Create a read response from the handler result."""
    event_time_timestamp = _timestamp_pb2.Timestamp()
    event_time_timestamp.FromDatetime(dt=response.event_time)
    result = source_pb2.ReadResponse.Result(
        payload=response.payload,
        keys=response.keys,
        offset=response.offset.as_dict,
        event_time=event_time_timestamp,
        headers=response.headers,
    )
    status = source_pb2.ReadResponse.Status(eot=False, code=source_pb2.ReadResponse.Status.SUCCESS)
    return source_pb2.ReadResponse(result=result, status=status)


def _create_eot_response():
    """Create an end-of-transmission response."""
    status = source_pb2.ReadResponse.Status(eot=True, code=source_pb2.ReadResponse.Status.SUCCESS)
    return source_pb2.ReadResponse(status=status)


def _create_ack_response():
    """Create an acknowledgement response."""
    return source_pb2.AckResponse(result=source_pb2.AckResponse.Result(success=_empty_pb2.Empty()))


class AsyncSourceServicer(source_pb2_grpc.SourceServicer):
    """
    This class is used to create a new grpc Source servicer instance.
    It implements the SourceServicer interface from the proto source.proto file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(self, source_handler: SourceCallable):
        self.background_tasks = set()
        self.source_handler = source_handler
        self.__initialize_handlers()
        self.cleanup_coroutines = []

    def __initialize_handlers(self):
        """Initialize handler methods from the provided source handler."""
        self.__source_read_handler = self.source_handler.read_handler
        self.__source_ack_handler = self.source_handler.ack_handler
        self.__source_nack_handler = self.source_handler.nack_handler
        self.__source_pending_handler = self.source_handler.pending_handler
        self.__source_partitions_handler = self.source_handler.partitions_handler

    async def ReadFn(
        self,
        request_iterator: AsyncIterator[source_pb2.ReadRequest],
        context: NumaflowServicerContext,
    ) -> AsyncIterator[source_pb2.ReadResponse]:
        """
        Handles the Read function, processing incoming requests and sending responses.
        """
        try:
            # The first message to be received should be a valid handshake
            req = await request_iterator.__anext__()
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise Exception("ReadFn: expected handshake message")
            yield _create_read_handshake_response()

            # process the incoming read requests on the stream
            async for req in request_iterator:
                # create an iterator to be provided to the user function where the responses will
                # be streamed
                niter = NonBlockingIterator()
                riter = niter.read_iterator()
                task = asyncio.create_task(self.__invoke_read(req, niter))
                # Save a reference to the result of this function, to avoid a
                # task disappearing mid-execution.
                self.background_tasks.add(task)
                task.add_done_callback(self.clean_background)

                async for resp in riter:
                    if isinstance(resp, BaseException):
                        await handle_async_error(context, resp, ERR_UDF_EXCEPTION_STRING)
                        return

                    yield _create_read_response(resp)

                # ensure that the task has completed
                await task
                # send an eot to signal all messages have been processed.
                yield _create_eot_response()
        except BaseException as err:
            _LOGGER.critical("User-Defined Source ReadFn error", exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)

    async def __invoke_read(self, req, niter):
        """Invoke the read handler and manage the iterator."""
        try:
            await self.__source_read_handler(
                ReadRequest(
                    num_records=req.request.num_records, timeout_in_ms=req.request.timeout_in_ms
                ),
                niter,
            )
            # Put an EOF to the iterator to indicate that we have completed the processing the
            # requests from the user handler
            await niter.put(STREAM_EOF)
        except BaseException as err:
            _LOGGER.critical("User-Defined Source ReadFn error", exc_info=True)
            await niter.put(err)

    async def AckFn(
        self,
        request_iterator: AsyncIterator[source_pb2.AckRequest],
        context: NumaflowServicerContext,
    ) -> AsyncIterator[source_pb2.AckResponse]:
        """
        Handles the Ack function for user-defined source.
        """
        try:
            # The first message to be received should be a valid handshake
            req = await request_iterator.__anext__()
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise Exception("AckFn: expected handshake message")
            yield _create_ack_handshake_response()

            # process the incoming Ack requests
            async for req in request_iterator:
                # proto repeated field(offsets) is of
                # type google._upb._message.RepeatedScalarContainer
                # we need to explicitly convert it to list
                offsets = [
                    Offset(offset.offset, offset.partition_id) for offset in req.request.offsets
                ]
                await self.__source_ack_handler(AckRequest(offsets=offsets))
                yield _create_ack_response()
        except BaseException as err:
            _LOGGER.critical("User-Defined Source AckFn error", exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)

    async def NackFn(
        self,
        request: source_pb2.NackRequest,
        context: NumaflowServicerContext,
    ) -> source_pb2.NackResponse:
        """
        Handles the Nack function for user-defined source.
        """
        try:
            offsets = [
                Offset(offset.offset, offset.partition_id) for offset in request.request.offsets
            ]
            await self.__source_nack_handler(NackRequest(offsets=offsets))
        except BaseException as err:
            _LOGGER.critical("User-Defined Source NackFn error", exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)
        return source_pb2.NackResponse(
            result=source_pb2.NackResponse.Result(success=_empty_pb2.Empty())
        )

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> source_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto source_pb2_grpc.py file.
        """
        return source_pb2.ReadyResponse(ready=True)

    async def PendingFn(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> source_pb2.PendingResponse:
        """
        PendingFn returns the number of pending records
        at the user defined source.
        """
        try:
            count = await self.__source_pending_handler()
        except BaseException as err:
            _LOGGER.critical("PendingFn Error", exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)
            return
        resp = source_pb2.PendingResponse.Result(count=count.count)
        return source_pb2.PendingResponse(result=resp)

    async def PartitionsFn(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> source_pb2.PartitionsResponse:
        """
        PartitionsFn returns the partitions of the user defined source.
        """
        try:
            partitions = await self.__source_partitions_handler()
        except BaseException as err:
            _LOGGER.critical("PartitionsFn Error", exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)
            return
        resp = source_pb2.PartitionsResponse.Result(partitions=partitions.partitions)
        return source_pb2.PartitionsResponse(result=resp)

    def clean_background(self, task):
        """
        Remove the task from the background tasks collection
        """
        self.background_tasks.remove(task)



import aiorun
import grpc

from pynumaflow._constants import (
    NUM_THREADS_DEFAULT,
    MAX_MESSAGE_SIZE,
    MAP_SOCK_PATH,
    MAP_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)
from pynumaflow.info.types import (
    ServerInfo,
    MAP_MODE_KEY,
    MapMode,
    MINIMUM_NUMAFLOW_VERSION,
    ContainerType,
)
from pynumaflow.mapper._dtypes import MapAsyncCallable
from pynumaflow.mapper._servicer._async_servicer import AsyncMapServicer
from pynumaflow.proto.mapper import map_pb2_grpc
from pynumaflow.shared.server import (
    NumaflowServer,
    start_async_server,
)


class MapAsyncServer(NumaflowServer):
    """
    Create a new grpc Map Server instance.
    Args:
        mapper_instance: The mapper instance to be used for Map UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16

    Example invocation:
        from pynumaflow.mapper import Messages, Message, Datum, MapAsyncServer
        async def async_map_handler(keys: list[str], datum: Datum) -> Messages:
            val = datum.value
            msg = "payload:{} event_time:{} watermark:{}".format(
                val.decode("utf-8"),
                datum.event_time,
                datum.watermark,
            )
            val = bytes(msg, encoding="utf-8")
            return Messages(Message(value=val, keys=keys))

        if __name__ == "__main__":
            grpc_server = MapAsyncServer(async_map_handler)
            grpc_server.start()
    """

    def __init__(
        self,
        mapper_instance: MapAsyncCallable,
        sock_path=MAP_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=MAP_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Asynchronous Map Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
        mapper_instance: The mapper instance to be used for Map UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to 4 and max capped at 16
        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.mapper_instance = mapper_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        # Get the servicer instance for the async server
        self.servicer = AsyncMapServicer(handler=mapper_instance)

    def start(self) -> None:
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self) -> None:
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """

        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context

        server_new = grpc.aio.server(options=self._server_options)
        server_new.add_insecure_port(self.sock_path)
        map_pb2_grpc.add_MapServicer_to_server(self.servicer, server_new)

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Mapper]
        # Add the MAP_MODE metadata to the server info for the correct map mode
        serv_info.metadata[MAP_MODE_KEY] = MapMode.UnaryMap

        # Start the async server
        await start_async_server(
            server_async=server_new,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



import inspect
from typing import Optional

import aiorun
import grpc

from pynumaflow.info.types import ServerInfo, ContainerType, MINIMUM_NUMAFLOW_VERSION
from pynumaflow.proto.reducer import reduce_pb2_grpc

from pynumaflow.reducestreamer.servicer.async_servicer import AsyncReduceStreamServicer

from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    REDUCE_STREAM_SOCK_PATH,
    REDUCE_STREAM_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.reducestreamer._dtypes import (
    ReduceStreamCallable,
    _ReduceStreamBuilderClass,
    ReduceStreamer,
)

from pynumaflow.shared.server import NumaflowServer, check_instance, start_async_server


def get_handler(
    reducer_handler: ReduceStreamCallable, init_args: tuple = (), init_kwargs: Optional[dict] = None
):
    """
    Get the correct handler type based on the arguments passed
    """
    if inspect.isfunction(reducer_handler):
        if init_args or init_kwargs:
            # if the init_args or init_kwargs are passed, then the reduce_stream_instance
            # can only be of class ReduceStreamer type
            raise TypeError("Cannot pass function handler with init args or kwargs")
        # return the function handler
        return reducer_handler
    elif not check_instance(reducer_handler, ReduceStreamer) and issubclass(
        reducer_handler, ReduceStreamer
    ):
        # if handler is type of Class ReduceStreamer, create a new instance of
        # a ReducerBuilderClass
        return _ReduceStreamBuilderClass(reducer_handler, init_args, init_kwargs)
    else:
        _LOGGER.error(
            _error_msg := f"Invalid Class Type {reducer_handler}: "
            f"Please make sure the class type is passed, and it is a subclass of ReduceStreamer"
        )
        raise TypeError(_error_msg)


class ReduceStreamAsyncServer(NumaflowServer):
    """
    Class for a new Reduce Stream Server instance.
    A new servicer instance is created and attached to the server.
    The server instance is returned.
    Args:
        reduce_stream_instance: The reducer instance to be used for
                Reduce Streaming UDF
        init_args: The arguments to be passed to the reduce_stream_handler
        init_kwargs: The keyword arguments to be passed to the
            reduce_stream_handler
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
        defaults to 4 and max capped at 16
        server_info_file: The path to the server info file
    Example invocation:
        import os
        from collections.abc import AsyncIterable
        from pynumaflow.reducestreamer import Messages, Message, Datum, Metadata,
        ReduceStreamAsyncServer, ReduceStreamer

        class ReduceCounter(ReduceStreamer):
            def __init__(self, counter):
                self.counter = counter

            async def handler(
                self,
                keys: list[str],
                datums: AsyncIterable[Datum],
                output: NonBlockingIterator,
                md: Metadata,
            ):
                async for _ in datums:
                    self.counter += 1
                    if self.counter > 20:
                        msg = f"counter:{self.counter}"
                        await output.put(Message(str.encode(msg), keys=keys))
                        self.counter = 0
                msg = f"counter:{self.counter}"
                await output.put(Message(str.encode(msg), keys=keys))

        async def reduce_handler(
                keys: list[str],
                datums: AsyncIterable[Datum],
                output: NonBlockingIterator,
                md: Metadata,
            ):
            counter = 0
            async for _ in datums:
                counter += 1
                if counter > 20:
                    msg = f"counter:{counter}"
                    await output.put(Message(str.encode(msg), keys=keys))
                    counter = 0
            msg = f"counter:{counter}"
            await output.put(Message(str.encode(msg), keys=keys))

        if __name__ == "__main__":
            invoke = os.getenv("INVOKE", "func_handler")
            if invoke == "class":
                # Here we are using the class instance as the reducer_instance
                # which will be used to invoke the handler function.
                # We are passing the init_args for the class instance.
                grpc_server = ReduceStreamAsyncServer(ReduceCounter, init_args=(0,))
            else:
                # Here we are using the handler function directly as the reducer_instance.
                grpc_server = ReduceStreamAsyncServer(reduce_handler)
            grpc_server.start()

    """

    def __init__(
        self,
        reduce_stream_instance: ReduceStreamCallable,
        init_args: tuple = (),
        init_kwargs: dict = None,
        sock_path=REDUCE_STREAM_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=REDUCE_STREAM_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Reduce Streamer Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            reduce_stream_instance: The reducer instance to be used for
                    Reduce Streaming UDF
            init_args: The arguments to be passed to the reduce_stream_handler
            init_kwargs: The keyword arguments to be passed to the
                reduce_stream_handler
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
            defaults to 4 and max capped at 16
            server_info_file: The path to the server info file
        """
        if init_kwargs is None:
            init_kwargs = {}
        self.reduce_stream_handler = get_handler(reduce_stream_instance, init_args, init_kwargs)
        self.sock_path = f"unix://{sock_path}"
        self.max_message_size = max_message_size
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.server_info_file = server_info_file

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        # Get the servicer instance for the async server
        self.servicer = AsyncReduceStreamServicer(self.reduce_stream_handler)

    def start(self):
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        _LOGGER.info(
            "Starting Async Reduce Stream Server",
        )
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new async server instance and add the servicer to it
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        reduce_pb2_grpc.add_ReduceServicer_to_server(self.servicer, server)

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Reducestreamer]
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



import asyncio
from collections.abc import AsyncIterable

from google.protobuf import empty_pb2 as _empty_pb2
from pynumaflow.shared.asynciter import NonBlockingIterator

from pynumaflow.shared.server import handle_async_error
from pynumaflow.sinker._dtypes import Datum, SinkAsyncCallable
from pynumaflow.proto.sinker import sink_pb2_grpc, sink_pb2
from pynumaflow.sinker.servicer.utils import (
    datum_from_sink_req,
    _create_read_handshake_response,
    build_sink_resp_results,
)
from pynumaflow.types import NumaflowServicerContext
from pynumaflow._constants import _LOGGER, STREAM_EOF, ERR_UDF_EXCEPTION_STRING


class AsyncSinkServicer(sink_pb2_grpc.SinkServicer):
    """
    This class is used to create a new grpc Sink servicer instance.
    It implements the SinkServicer interface from the proto sink.proto file.
    Provides the functionality for the required rpc methods.
    """

    def __init__(
        self,
        handler: SinkAsyncCallable,
    ):
        self.background_tasks = set()
        self.__sink_handler: SinkAsyncCallable = handler
        self.cleanup_coroutines = []

    async def SinkFn(
        self,
        request_iterator: AsyncIterable[sink_pb2.SinkRequest],
        context: NumaflowServicerContext,
    ) -> sink_pb2.SinkResponse:
        """
        Applies a sink function to a list of datum elements.
        The pascal case function name comes from the proto sink_pb2_grpc.py file.
        """
        try:
            # The first message to be received should be a valid handshake
            req = await request_iterator.__anext__()
            # check if it is a valid handshake req
            if not (req.handshake and req.handshake.sot):
                raise Exception("ReadFn: expected handshake message")
            yield _create_read_handshake_response()

            # cur_task is used to track the task (coroutine) processing
            # the current batch of messages.
            cur_task = None
            # iterate of the incoming messages ot the sink
            async for d in request_iterator:
                # if we do not have any active task currently processing the batch
                # we need to create one and call the User function for processing the same.
                if cur_task is None:
                    req_queue = NonBlockingIterator()
                    cur_task = asyncio.create_task(
                        self.__invoke_sink(req_queue.read_iterator(), context)
                    )
                    self.background_tasks.add(cur_task)
                    cur_task.add_done_callback(self.background_tasks.discard)

                # when we have end of transmission message, we need to stop the processing the
                # current batch and wait for the next batch of messages.
                # We will also wait for the current task to finish processing the current batch.
                # We mark the current task as None to indicate that we are
                # ready to process the next batch.
                if d.status and d.status.eot:
                    await req_queue.put(STREAM_EOF)
                    await cur_task
                    ret = cur_task.result()
                    yield sink_pb2.SinkResponse(results=ret)
                    # send EOT after each finishing sink responses
                    yield sink_pb2.SinkResponse(status=sink_pb2.TransmissionStatus(eot=True))
                    cur_task = None
                    continue

                # if we have a valid message, we will add it to the request queue for processing.
                datum = datum_from_sink_req(d)
                await req_queue.put(datum)
        except BaseException as err:
            # if there is an exception, we will mark all the responses as a failure
            err_msg = f"UDSinkError: {repr(err)}"
            _LOGGER.critical(err_msg, exc_info=True)
            await handle_async_error(context, err, ERR_UDF_EXCEPTION_STRING)
            return

    async def __invoke_sink(
        self, request_queue: AsyncIterable[Datum], context: NumaflowServicerContext
    ):
        try:
            # invoke the user function with the request queue
            rspns = await self.__sink_handler(request_queue)
            return build_sink_resp_results(rspns)
        except BaseException as err:
            err_msg = f"UDSinkError: {repr(err)}"
            _LOGGER.critical(err_msg, exc_info=True)
            raise err

    async def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> sink_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto sink_pb2_grpc.py file.
        """
        return sink_pb2.ReadyResponse(ready=True)



from pynumaflow.accumulator._dtypes import (
    Message,
    Datum,
    IntervalWindow,
    DROP,
    KeyedWindow,
    Accumulator,
)
from pynumaflow.accumulator.async_server import AccumulatorAsyncServer

__all__ = [
    "Message",
    "Datum",
    "IntervalWindow",
    "DROP",
    "AccumulatorAsyncServer",
    "KeyedWindow",
    "Accumulator",
]



import aiorun
import grpc

from pynumaflow.info.types import ServerInfo, ContainerType, MINIMUM_NUMAFLOW_VERSION
from pynumaflow.sourcer.servicer.async_servicer import AsyncSourceServicer

from pynumaflow._constants import (
    SOURCE_SOCK_PATH,
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    SOURCE_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)
from pynumaflow.proto.sourcer import source_pb2_grpc

from pynumaflow.shared.server import NumaflowServer, start_async_server
from pynumaflow.sourcer._dtypes import SourceCallable


class SourceAsyncServer(NumaflowServer):
    """
    Class for a new Async Source Server instance.
    """

    def __init__(
        self,
        sourcer_instance: SourceCallable,
        sock_path=SOURCE_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=SOURCE_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Async Source Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            sourcer_instance: The sourcer instance to be used for Source UDF
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                            defaults to 4 and max capped at 16

        Example invocation:
            from datetime import datetime
            from pynumaflow.shared.asynciter import NonBlockingIterator
            from pynumaflow.sourcer import (
                ReadRequest,
                Message,
                AckRequest,
                PendingResponse,
                Offset,
                PartitionsResponse,
                get_default_partitions,
                Sourcer,
                SourceAsyncServer,
                NackRequest,
            )

            class AsyncSource(Sourcer):
                # AsyncSource is a class for User Defined Source implementation.

                def __init__(self):
                    # The offset idx till where the messages have been read
                    self.read_idx: int = 0
                    # Set to maintain a track of the offsets yet to be acknowledged
                    self.to_ack_set: set[int] = set()
                    # Set to maintain a track of the offsets that have been negatively acknowledged
                    self.nacked: set[int] = set()

                async def read_handler(self, datum: ReadRequest, output: NonBlockingIterator):
                    '''
                    read_handler is used to read the data from the source and send the data forward
                    for each read request we process num_records and increment the read_idx to
                    indicate that the message has been read and the same is added to the ack set
                    '''
                    if self.to_ack_set:
                        return

                    for x in range(datum.num_records):
                        # If there are any nacked offsets, re-deliver them
                        if self.nacked:
                            idx = self.nacked.pop()
                        else:
                            idx = self.read_idx
                            self.read_idx += 1
                        headers = {"x-txn-id": str(uuid.uuid4())}
                        await output.put(
                            Message(
                                payload=str(self.read_idx).encode(),
                                offset=Offset.offset_with_default_partition_id(str(idx).encode()),
                                event_time=datetime.now(),
                                headers=headers,
                            )
                        )
                        self.to_ack_set.add(idx)

                async def ack_handler(self, ack_request: AckRequest):
                    '''
                    The ack handler is used acknowledge the offsets that have been read, and remove
                    them from the to_ack_set
                    '''
                    for req in ack_request.offsets:
                        offset = int(req.offset)
                        self.to_ack_set.remove(offset)

                async def nack_handler(self, ack_request: NackRequest):
                    '''
                    Add the offsets that have been negatively acknowledged to the nacked set
                    '''
                    for req in ack_request.offsets:
                        offset = int(req.offset)
                        self.to_ack_set.remove(offset)
                        self.nacked.add(offset)

                async def pending_handler(self) -> PendingResponse:
                    '''
                    The simple source always returns zero to indicate there is no pending record.
                    '''
                    return PendingResponse(count=0)

                async def partitions_handler(self) -> PartitionsResponse:
                    '''
                    The simple source always returns default partitions.
                    '''
                    return PartitionsResponse(partitions=get_default_partitions())


            if __name__ == "__main__":
                ud_source = AsyncSource()
                grpc_server = SourceAsyncServer(ud_source)
                grpc_server.start()

        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.sourcer_instance = sourcer_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]

        self.servicer = AsyncSourceServicer(source_handler=sourcer_instance)

    def start(self):
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Async gRPC server on the given UNIX socket with given max threads
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new async server instance and add the servicer to it
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        source_servicer = self.servicer
        source_pb2_grpc.add_SourceServicer_to_server(source_servicer, server)

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Sourcer]
        # Start the async server
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



from pynumaflow.info.types import ServerInfo, MINIMUM_NUMAFLOW_VERSION, ContainerType
from pynumaflow.shared import NumaflowServer
from pynumaflow.shared.server import sync_server_start
from pynumaflow.sideinput._dtypes import RetrieverCallable
from pynumaflow.sideinput.servicer.servicer import SideInputServicer
from pynumaflow._constants import (
    NUM_THREADS_DEFAULT,
    MAX_MESSAGE_SIZE,
    SIDE_INPUT_SOCK_PATH,
    _LOGGER,
    UDFType,
    SIDE_INPUT_DIR_PATH,
    SIDE_INPUT_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)


class SideInputServer(NumaflowServer):
    """
    Class for a new Side Input Server instance.
    Args:
        side_input_instance: The side input instance to be used for Side Input UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;

    Example invocation:
        import datetime
        from pynumaflow.sideinput import Response, SideInputServer, SideInput

        class ExampleSideInput(SideInput):
            def __init__(self):
                self.counter = 0

            def retrieve_handler(self) -> Response:
                time_now = datetime.datetime.now()
                # val is the value to be broadcasted
                val = f"an example: {str(time_now)}"
                self.counter += 1
                # broadcast every other time
                if self.counter % 2 == 0:
                    # no_broadcast_message() is used to indicate that there is no broadcast
                    return Response.no_broadcast_message()
                # broadcast_message() is used to indicate that there is a broadcast
                return Response.broadcast_message(val.encode("utf-8"))

        if __name__ == "__main__":
            grpc_server = SideInputServer(ExampleSideInput())
            grpc_server.start()

    """

    def __init__(
        self,
        side_input_instance: RetrieverCallable,
        sock_path=SIDE_INPUT_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        side_input_dir_path=SIDE_INPUT_DIR_PATH,
        server_info_file=SIDE_INPUT_SERVER_INFO_FILE_PATH,
    ):
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]

        self.side_input_instance = side_input_instance
        self.side_input_dir_path = side_input_dir_path
        self.servicer = SideInputServicer(side_input_instance)

    def start(self):
        """
        Starts the Synchronous gRPC server on the given UNIX socket with given max threads.
        """
        # Get the servicer instance based on the server type
        side_input_servicer = self.servicer

        _LOGGER.info(
            "Side Input GRPC Server listening on: %s with max threads: %s",
            self.sock_path,
            self.max_threads,
        )

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Sideinput]
        # Start the server
        sync_server_start(
            servicer=side_input_servicer,
            bind_address=self.sock_path,
            max_threads=self.max_threads,
            server_info_file=self.server_info_file,
            server_options=self._server_options,
            udf_type=UDFType.SideInput,
            server_info=serv_info,
        )



from pynumaflow.sourcetransformer._dtypes import (
    Message,
    Messages,
    Datum,
    DROP,
    SourceTransformer,
)
from pynumaflow.sourcetransformer.multiproc_server import SourceTransformMultiProcServer
from pynumaflow.sourcetransformer.server import SourceTransformServer
from pynumaflow.sourcetransformer.async_server import SourceTransformAsyncServer

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "SourceTransformServer",
    "SourceTransformer",
    "SourceTransformMultiProcServer",
    "SourceTransformAsyncServer",
]



import os
import json
import threading
import time
from pynumaflow._constants import (
    CONTAINER_TYPE,
    RUNTIME_APPLICATION_ERRORS_PATH,
    CURRENT_CRITICAL_ERROR_FILE,
    INTERNAL_ERROR_CODE,
)
from pynumaflow.errors._dtypes import _RuntimeErrorEntry
from typing import Union


class _PersistErrorOnce:
    """Ensures that the persist_critical_error function is executed only once."""

    def __init__(self):
        self.done = False
        self.lock = threading.Lock()

    def execute(self, func, *args, **kwargs):
        with self.lock:
            if self.done:
                raise RuntimeError("Persist critical error function has already been executed.")
            self.done = True
            return func(*args, **kwargs)


_persist_error_once = _PersistErrorOnce()


def persist_critical_error(
    error_code: str, error_message: str, error_details: str
) -> Union[RuntimeError, None]:
    """
    Persists a critical error to a file. This function will only execute once.
    Logs the error if persisting to the file fails.
    Returns None if successful, or raises RuntimeError if already executed.
    """
    try:
        _persist_error_once.execute(
            _persist_critical_error_to_file,
            error_code,
            error_message,
            error_details,
            RUNTIME_APPLICATION_ERRORS_PATH,
        )
    except RuntimeError as e:
        return e
    except Exception as e:
        print(f"Error in persisting critical error: {e}")
    return None


def _persist_critical_error_to_file(
    error_code: str, error_message: str, error_details: str, dir_path: str
):
    """Internal function to persist a critical error to a file."""

    os.makedirs(dir_path, mode=0o777, exist_ok=True)
    container_dir = os.path.join(dir_path, CONTAINER_TYPE)
    os.makedirs(container_dir, mode=0o777, exist_ok=True)

    current_file_path = os.path.join(container_dir, CURRENT_CRITICAL_ERROR_FILE)
    error_code = error_code or INTERNAL_ERROR_CODE
    current_timestamp = int(time.time())

    runtime_error_entry = _RuntimeErrorEntry(
        container=CONTAINER_TYPE,
        timestamp=current_timestamp,
        code=error_code,
        message=error_message,
        details=error_details,
    )

    with open(current_file_path, "w") as f:
        json.dump(runtime_error_entry.to_dict(), f)

    final_file_name = f"{current_timestamp}-udf.json"
    final_file_path = os.path.join(container_dir, final_file_name)
    os.rename(current_file_path, final_file_path)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/mapper/map.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/mapper/map.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!pynumaflow/proto/mapper/map.proto\x12\x06map.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xac\x03\n\nMapRequest\x12+\n\x07request\x18\x01 \x01(\x0b\x32\x1a.map.v1.MapRequest.Request\x12\n\n\x02id\x18\x02 \x01(\t\x12)\n\thandshake\x18\x03 \x01(\x0b\x32\x11.map.v1.HandshakeH\x00\x88\x01\x01\x12/\n\x06status\x18\x04 \x01(\x0b\x32\x1a.map.v1.TransmissionStatusH\x01\x88\x01\x01\x1a\xef\x01\n\x07Request\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x38\n\x07headers\x18\x05 \x03(\x0b\x32\'.map.v1.MapRequest.Request.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0c\n\n_handshakeB\t\n\x07_status\"\x18\n\tHandshake\x12\x0b\n\x03sot\x18\x01 \x01(\x08\"!\n\x12TransmissionStatus\x12\x0b\n\x03\x65ot\x18\x01 \x01(\x08\"\xf0\x01\n\x0bMapResponse\x12+\n\x07results\x18\x01 \x03(\x0b\x32\x1a.map.v1.MapResponse.Result\x12\n\n\x02id\x18\x02 \x01(\t\x12)\n\thandshake\x18\x03 \x01(\x0b\x32\x11.map.v1.HandshakeH\x00\x88\x01\x01\x12/\n\x06status\x18\x04 \x01(\x0b\x32\x1a.map.v1.TransmissionStatusH\x01\x88\x01\x01\x1a\x33\n\x06Result\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x0c\n\x04tags\x18\x03 \x03(\tB\x0c\n\n_handshakeB\t\n\x07_status\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32u\n\x03Map\x12\x34\n\x05MapFn\x12\x12.map.v1.MapRequest\x1a\x13.map.v1.MapResponse(\x01\x30\x01\x12\x38\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x15.map.v1.ReadyResponseB7Z5github.com/numaproj/numaflow-go/pkg/apis/proto/map/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.mapper.map_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z5github.com/numaproj/numaflow-go/pkg/apis/proto/map/v1'
  _globals['_MAPREQUEST_REQUEST_HEADERSENTRY']._loaded_options = None
  _globals['_MAPREQUEST_REQUEST_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_MAPREQUEST']._serialized_start=108
  _globals['_MAPREQUEST']._serialized_end=536
  _globals['_MAPREQUEST_REQUEST']._serialized_start=272
  _globals['_MAPREQUEST_REQUEST']._serialized_end=511
  _globals['_MAPREQUEST_REQUEST_HEADERSENTRY']._serialized_start=465
  _globals['_MAPREQUEST_REQUEST_HEADERSENTRY']._serialized_end=511
  _globals['_HANDSHAKE']._serialized_start=538
  _globals['_HANDSHAKE']._serialized_end=562
  _globals['_TRANSMISSIONSTATUS']._serialized_start=564
  _globals['_TRANSMISSIONSTATUS']._serialized_end=597
  _globals['_MAPRESPONSE']._serialized_start=600
  _globals['_MAPRESPONSE']._serialized_end=840
  _globals['_MAPRESPONSE_RESULT']._serialized_start=764
  _globals['_MAPRESPONSE_RESULT']._serialized_end=815
  _globals['_READYRESPONSE']._serialized_start=842
  _globals['_READYRESPONSE']._serialized_end=872
  _globals['_MAP']._serialized_start=874
  _globals['_MAP']._serialized_end=991
# @@protoc_insertion_point(module_scope)



from pynumaflow._constants import DROP

from pynumaflow.batchmapper._dtypes import (
    Message,
    Datum,
    BatchMapper,
    BatchResponses,
    BatchResponse,
)
from pynumaflow.batchmapper.async_server import BatchMapAsyncServer

__all__ = [
    "Message",
    "Datum",
    "DROP",
    "BatchMapAsyncServer",
    "BatchMapper",
    "BatchResponses",
    "BatchResponse",
]



import os

import aiorun
import grpc

from pynumaflow.info.types import ContainerType, ServerInfo, MINIMUM_NUMAFLOW_VERSION
from pynumaflow.sinker.servicer.async_servicer import AsyncSinkServicer
from pynumaflow.proto.sinker import sink_pb2_grpc


from pynumaflow._constants import (
    SINK_SOCK_PATH,
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    SINK_SERVER_INFO_FILE_PATH,
    ENV_UD_CONTAINER_TYPE,
    UD_CONTAINER_FALLBACK_SINK,
    _LOGGER,
    FALLBACK_SINK_SOCK_PATH,
    FALLBACK_SINK_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.shared.server import NumaflowServer, start_async_server
from pynumaflow.sinker._dtypes import SinkAsyncCallable


class SinkAsyncServer(NumaflowServer):
    """
    SinkAsyncServer is the main class to start a gRPC server for a sinker.
    Create a new grpc Async Sink Server instance.
    A new servicer instance is created and attached to the server.
    The server instance is returned.
    Args:
        sinker_instance: The sinker instance to be used for Sink UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16

    Example invocation:
        import os
        from collections.abc import AsyncIterable
        from pynumaflow.sinker import Datum, Responses, Response, Sinker
        from pynumaflow.sinker import SinkAsyncServer
        from pynumaflow._constants import _LOGGER


        class UserDefinedSink(Sinker):
            async def handler(self, datums: AsyncIterable[Datum]) -> Responses:
                responses = Responses()
                async for msg in datums:
                    _LOGGER.info("User Defined Sink %s", msg.value.decode("utf-8"))
                    responses.append(Response.as_success(msg.id))
                return responses


        async def udsink_handler(datums: AsyncIterable[Datum]) -> Responses:
            responses = Responses()
            async for msg in datums:
                _LOGGER.info("User Defined Sink %s", msg.value.decode("utf-8"))
                responses.append(Response.as_success(msg.id))
            return responses


        if __name__ == "__main__":
            invoke = os.getenv("INVOKE", "func_handler")
            if invoke == "class":
                sink_handler = UserDefinedSink()
            else:
                sink_handler = udsink_handler
            grpc_server = SinkAsyncServer(sink_handler)
            grpc_server.start()
    """

    def __init__(
        self,
        sinker_instance: SinkAsyncCallable,
        sock_path=SINK_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=SINK_SERVER_INFO_FILE_PATH,
    ):
        # If the container type is fallback sink, then use the fallback sink address and path.
        if os.getenv(ENV_UD_CONTAINER_TYPE, "") == UD_CONTAINER_FALLBACK_SINK:
            _LOGGER.info("Using Fallback Sink")
            sock_path = FALLBACK_SINK_SOCK_PATH
            server_info_file = FALLBACK_SINK_SERVER_INFO_FILE_PATH

        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.sinker_instance = sinker_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]

        self.servicer = AsyncSinkServicer(sinker_instance)

    def start(self):
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Asynchronous gRPC server on the given UNIX socket with given max threads.
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new server instance, add the servicer to it and start the server
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        sink_pb2_grpc.add_SinkServicer_to_server(self.servicer, server)
        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Sinker]
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/sourcer/source.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/sourcer/source.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%pynumaflow/proto/sourcer/source.proto\x12\tsource.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto\"\x18\n\tHandshake\x12\x0b\n\x03sot\x18\x01 \x01(\x08\"\xb1\x01\n\x0bReadRequest\x12/\n\x07request\x18\x01 \x01(\x0b\x32\x1e.source.v1.ReadRequest.Request\x12,\n\thandshake\x18\x02 \x01(\x0b\x32\x14.source.v1.HandshakeH\x00\x88\x01\x01\x1a\x35\n\x07Request\x12\x13\n\x0bnum_records\x18\x01 \x01(\x04\x12\x15\n\rtimeout_in_ms\x18\x02 \x01(\rB\x0c\n\n_handshake\"\x81\x05\n\x0cReadResponse\x12.\n\x06result\x18\x01 \x01(\x0b\x32\x1e.source.v1.ReadResponse.Result\x12.\n\x06status\x18\x02 \x01(\x0b\x32\x1e.source.v1.ReadResponse.Status\x12,\n\thandshake\x18\x03 \x01(\x0b\x32\x14.source.v1.HandshakeH\x00\x88\x01\x01\x1a\xe8\x01\n\x06Result\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x12!\n\x06offset\x18\x02 \x01(\x0b\x32\x11.source.v1.Offset\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04keys\x18\x04 \x03(\t\x12<\n\x07headers\x18\x05 \x03(\x0b\x32+.source.v1.ReadResponse.Result.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xe9\x01\n\x06Status\x12\x0b\n\x03\x65ot\x18\x01 \x01(\x08\x12\x31\n\x04\x63ode\x18\x02 \x01(\x0e\x32#.source.v1.ReadResponse.Status.Code\x12\x38\n\x05\x65rror\x18\x03 \x01(\x0e\x32$.source.v1.ReadResponse.Status.ErrorH\x00\x88\x01\x01\x12\x10\n\x03msg\x18\x04 \x01(\tH\x01\x88\x01\x01\" \n\x04\x43ode\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\"\x1f\n\x05\x45rror\x12\x0b\n\x07UNACKED\x10\x00\x12\t\n\x05OTHER\x10\x01\x42\x08\n\x06_errorB\x06\n\x04_msgB\x0c\n\n_handshake\"\xa7\x01\n\nAckRequest\x12.\n\x07request\x18\x01 \x01(\x0b\x32\x1d.source.v1.AckRequest.Request\x12,\n\thandshake\x18\x02 \x01(\x0b\x32\x14.source.v1.HandshakeH\x00\x88\x01\x01\x1a-\n\x07Request\x12\"\n\x07offsets\x18\x01 \x03(\x0b\x32\x11.source.v1.OffsetB\x0c\n\n_handshake\"\xab\x01\n\x0b\x41\x63kResponse\x12-\n\x06result\x18\x01 \x01(\x0b\x32\x1d.source.v1.AckResponse.Result\x12,\n\thandshake\x18\x02 \x01(\x0b\x32\x14.source.v1.HandshakeH\x00\x88\x01\x01\x1a\x31\n\x06Result\x12\'\n\x07success\x18\x01 \x01(\x0b\x32\x16.google.protobuf.EmptyB\x0c\n\n_handshake\"m\n\x0bNackRequest\x12/\n\x07request\x18\x01 \x01(\x0b\x32\x1e.source.v1.NackRequest.Request\x1a-\n\x07Request\x12\"\n\x07offsets\x18\x01 \x03(\x0b\x32\x11.source.v1.Offset\"q\n\x0cNackResponse\x12.\n\x06result\x18\x01 \x01(\x0b\x32\x1e.source.v1.NackResponse.Result\x1a\x31\n\x06Result\x12\'\n\x07success\x18\x01 \x01(\x0b\x32\x16.google.protobuf.Empty\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\"]\n\x0fPendingResponse\x12\x31\n\x06result\x18\x01 \x01(\x0b\x32!.source.v1.PendingResponse.Result\x1a\x17\n\x06Result\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\"h\n\x12PartitionsResponse\x12\x34\n\x06result\x18\x01 \x01(\x0b\x32$.source.v1.PartitionsResponse.Result\x1a\x1c\n\x06Result\x12\x12\n\npartitions\x18\x01 \x03(\x05\".\n\x06Offset\x12\x0e\n\x06offset\x18\x01 \x01(\x0c\x12\x14\n\x0cpartition_id\x18\x02 \x01(\x05\x32\x83\x03\n\x06Source\x12=\n\x06ReadFn\x12\x16.source.v1.ReadRequest\x1a\x17.source.v1.ReadResponse(\x01\x30\x01\x12:\n\x05\x41\x63kFn\x12\x15.source.v1.AckRequest\x1a\x16.source.v1.AckResponse(\x01\x30\x01\x12\x39\n\x06NackFn\x12\x16.source.v1.NackRequest\x1a\x17.source.v1.NackResponse\x12?\n\tPendingFn\x12\x16.google.protobuf.Empty\x1a\x1a.source.v1.PendingResponse\x12\x45\n\x0cPartitionsFn\x12\x16.google.protobuf.Empty\x1a\x1d.source.v1.PartitionsResponse\x12;\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x18.source.v1.ReadyResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.sourcer.source_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_READRESPONSE_RESULT_HEADERSENTRY']._loaded_options = None
  _globals['_READRESPONSE_RESULT_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_HANDSHAKE']._serialized_start=114
  _globals['_HANDSHAKE']._serialized_end=138
  _globals['_READREQUEST']._serialized_start=141
  _globals['_READREQUEST']._serialized_end=318
  _globals['_READREQUEST_REQUEST']._serialized_start=251
  _globals['_READREQUEST_REQUEST']._serialized_end=304
  _globals['_READRESPONSE']._serialized_start=321
  _globals['_READRESPONSE']._serialized_end=962
  _globals['_READRESPONSE_RESULT']._serialized_start=480
  _globals['_READRESPONSE_RESULT']._serialized_end=712
  _globals['_READRESPONSE_RESULT_HEADERSENTRY']._serialized_start=666
  _globals['_READRESPONSE_RESULT_HEADERSENTRY']._serialized_end=712
  _globals['_READRESPONSE_STATUS']._serialized_start=715
  _globals['_READRESPONSE_STATUS']._serialized_end=948
  _globals['_READRESPONSE_STATUS_CODE']._serialized_start=865
  _globals['_READRESPONSE_STATUS_CODE']._serialized_end=897
  _globals['_READRESPONSE_STATUS_ERROR']._serialized_start=899
  _globals['_READRESPONSE_STATUS_ERROR']._serialized_end=930
  _globals['_ACKREQUEST']._serialized_start=965
  _globals['_ACKREQUEST']._serialized_end=1132
  _globals['_ACKREQUEST_REQUEST']._serialized_start=1073
  _globals['_ACKREQUEST_REQUEST']._serialized_end=1118
  _globals['_ACKRESPONSE']._serialized_start=1135
  _globals['_ACKRESPONSE']._serialized_end=1306
  _globals['_ACKRESPONSE_RESULT']._serialized_start=1243
  _globals['_ACKRESPONSE_RESULT']._serialized_end=1292
  _globals['_NACKREQUEST']._serialized_start=1308
  _globals['_NACKREQUEST']._serialized_end=1417
  _globals['_NACKREQUEST_REQUEST']._serialized_start=1073
  _globals['_NACKREQUEST_REQUEST']._serialized_end=1118
  _globals['_NACKRESPONSE']._serialized_start=1419
  _globals['_NACKRESPONSE']._serialized_end=1532
  _globals['_NACKRESPONSE_RESULT']._serialized_start=1243
  _globals['_NACKRESPONSE_RESULT']._serialized_end=1292
  _globals['_READYRESPONSE']._serialized_start=1534
  _globals['_READYRESPONSE']._serialized_end=1564
  _globals['_PENDINGRESPONSE']._serialized_start=1566
  _globals['_PENDINGRESPONSE']._serialized_end=1659
  _globals['_PENDINGRESPONSE_RESULT']._serialized_start=1636
  _globals['_PENDINGRESPONSE_RESULT']._serialized_end=1659
  _globals['_PARTITIONSRESPONSE']._serialized_start=1661
  _globals['_PARTITIONSRESPONSE']._serialized_end=1765
  _globals['_PARTITIONSRESPONSE_RESULT']._serialized_start=1737
  _globals['_PARTITIONSRESPONSE_RESULT']._serialized_end=1765
  _globals['_OFFSET']._serialized_start=1767
  _globals['_OFFSET']._serialized_end=1813
  _globals['_SOURCE']._serialized_start=1816
  _globals['_SOURCE']._serialized_end=2203
# @@protoc_insertion_point(module_scope)



import datetime

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import empty_pb2 as _empty_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Handshake(_message.Message):
    __slots__ = ("sot",)
    SOT_FIELD_NUMBER: _ClassVar[int]
    sot: bool
    def __init__(self, sot: bool = ...) -> None: ...

class ReadRequest(_message.Message):
    __slots__ = ("request", "handshake")
    class Request(_message.Message):
        __slots__ = ("num_records", "timeout_in_ms")
        NUM_RECORDS_FIELD_NUMBER: _ClassVar[int]
        TIMEOUT_IN_MS_FIELD_NUMBER: _ClassVar[int]
        num_records: int
        timeout_in_ms: int
        def __init__(self, num_records: _Optional[int] = ..., timeout_in_ms: _Optional[int] = ...) -> None: ...
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    HANDSHAKE_FIELD_NUMBER: _ClassVar[int]
    request: ReadRequest.Request
    handshake: Handshake
    def __init__(self, request: _Optional[_Union[ReadRequest.Request, _Mapping]] = ..., handshake: _Optional[_Union[Handshake, _Mapping]] = ...) -> None: ...

class ReadResponse(_message.Message):
    __slots__ = ("result", "status", "handshake")
    class Result(_message.Message):
        __slots__ = ("payload", "offset", "event_time", "keys", "headers")
        class HeadersEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: str
            def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
        PAYLOAD_FIELD_NUMBER: _ClassVar[int]
        OFFSET_FIELD_NUMBER: _ClassVar[int]
        EVENT_TIME_FIELD_NUMBER: _ClassVar[int]
        KEYS_FIELD_NUMBER: _ClassVar[int]
        HEADERS_FIELD_NUMBER: _ClassVar[int]
        payload: bytes
        offset: Offset
        event_time: _timestamp_pb2.Timestamp
        keys: _containers.RepeatedScalarFieldContainer[str]
        headers: _containers.ScalarMap[str, str]
        def __init__(self, payload: _Optional[bytes] = ..., offset: _Optional[_Union[Offset, _Mapping]] = ..., event_time: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., keys: _Optional[_Iterable[str]] = ..., headers: _Optional[_Mapping[str, str]] = ...) -> None: ...
    class Status(_message.Message):
        __slots__ = ("eot", "code", "error", "msg")
        class Code(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            SUCCESS: _ClassVar[ReadResponse.Status.Code]
            FAILURE: _ClassVar[ReadResponse.Status.Code]
        SUCCESS: ReadResponse.Status.Code
        FAILURE: ReadResponse.Status.Code
        class Error(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            UNACKED: _ClassVar[ReadResponse.Status.Error]
            OTHER: _ClassVar[ReadResponse.Status.Error]
        UNACKED: ReadResponse.Status.Error
        OTHER: ReadResponse.Status.Error
        EOT_FIELD_NUMBER: _ClassVar[int]
        CODE_FIELD_NUMBER: _ClassVar[int]
        ERROR_FIELD_NUMBER: _ClassVar[int]
        MSG_FIELD_NUMBER: _ClassVar[int]
        eot: bool
        code: ReadResponse.Status.Code
        error: ReadResponse.Status.Error
        msg: str
        def __init__(self, eot: bool = ..., code: _Optional[_Union[ReadResponse.Status.Code, str]] = ..., error: _Optional[_Union[ReadResponse.Status.Error, str]] = ..., msg: _Optional[str] = ...) -> None: ...
    RESULT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    HANDSHAKE_FIELD_NUMBER: _ClassVar[int]
    result: ReadResponse.Result
    status: ReadResponse.Status
    handshake: Handshake
    def __init__(self, result: _Optional[_Union[ReadResponse.Result, _Mapping]] = ..., status: _Optional[_Union[ReadResponse.Status, _Mapping]] = ..., handshake: _Optional[_Union[Handshake, _Mapping]] = ...) -> None: ...

class AckRequest(_message.Message):
    __slots__ = ("request", "handshake")
    class Request(_message.Message):
        __slots__ = ("offsets",)
        OFFSETS_FIELD_NUMBER: _ClassVar[int]
        offsets: _containers.RepeatedCompositeFieldContainer[Offset]
        def __init__(self, offsets: _Optional[_Iterable[_Union[Offset, _Mapping]]] = ...) -> None: ...
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    HANDSHAKE_FIELD_NUMBER: _ClassVar[int]
    request: AckRequest.Request
    handshake: Handshake
    def __init__(self, request: _Optional[_Union[AckRequest.Request, _Mapping]] = ..., handshake: _Optional[_Union[Handshake, _Mapping]] = ...) -> None: ...

class AckResponse(_message.Message):
    __slots__ = ("result", "handshake")
    class Result(_message.Message):
        __slots__ = ("success",)
        SUCCESS_FIELD_NUMBER: _ClassVar[int]
        success: _empty_pb2.Empty
        def __init__(self, success: _Optional[_Union[_empty_pb2.Empty, _Mapping]] = ...) -> None: ...
    RESULT_FIELD_NUMBER: _ClassVar[int]
    HANDSHAKE_FIELD_NUMBER: _ClassVar[int]
    result: AckResponse.Result
    handshake: Handshake
    def __init__(self, result: _Optional[_Union[AckResponse.Result, _Mapping]] = ..., handshake: _Optional[_Union[Handshake, _Mapping]] = ...) -> None: ...

class NackRequest(_message.Message):
    __slots__ = ("request",)
    class Request(_message.Message):
        __slots__ = ("offsets",)
        OFFSETS_FIELD_NUMBER: _ClassVar[int]
        offsets: _containers.RepeatedCompositeFieldContainer[Offset]
        def __init__(self, offsets: _Optional[_Iterable[_Union[Offset, _Mapping]]] = ...) -> None: ...
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    request: NackRequest.Request
    def __init__(self, request: _Optional[_Union[NackRequest.Request, _Mapping]] = ...) -> None: ...

class NackResponse(_message.Message):
    __slots__ = ("result",)
    class Result(_message.Message):
        __slots__ = ("success",)
        SUCCESS_FIELD_NUMBER: _ClassVar[int]
        success: _empty_pb2.Empty
        def __init__(self, success: _Optional[_Union[_empty_pb2.Empty, _Mapping]] = ...) -> None: ...
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: NackResponse.Result
    def __init__(self, result: _Optional[_Union[NackResponse.Result, _Mapping]] = ...) -> None: ...

class ReadyResponse(_message.Message):
    __slots__ = ("ready",)
    READY_FIELD_NUMBER: _ClassVar[int]
    ready: bool
    def __init__(self, ready: bool = ...) -> None: ...

class PendingResponse(_message.Message):
    __slots__ = ("result",)
    class Result(_message.Message):
        __slots__ = ("count",)
        COUNT_FIELD_NUMBER: _ClassVar[int]
        count: int
        def __init__(self, count: _Optional[int] = ...) -> None: ...
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: PendingResponse.Result
    def __init__(self, result: _Optional[_Union[PendingResponse.Result, _Mapping]] = ...) -> None: ...

class PartitionsResponse(_message.Message):
    __slots__ = ("result",)
    class Result(_message.Message):
        __slots__ = ("partitions",)
        PARTITIONS_FIELD_NUMBER: _ClassVar[int]
        partitions: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, partitions: _Optional[_Iterable[int]] = ...) -> None: ...
    RESULT_FIELD_NUMBER: _ClassVar[int]
    result: PartitionsResponse.Result
    def __init__(self, result: _Optional[_Union[PartitionsResponse.Result, _Mapping]] = ...) -> None: ...

class Offset(_message.Message):
    __slots__ = ("offset", "partition_id")
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    PARTITION_ID_FIELD_NUMBER: _ClassVar[int]
    offset: bytes
    partition_id: int
    def __init__(self, offset: _Optional[bytes] = ..., partition_id: _Optional[int] = ...) -> None: ...



from pynumaflow.proto.sinker import sink_pb2
from pynumaflow.sinker._dtypes import Response, Datum, Responses


def build_sink_resp_results(responses: Responses) -> list[sink_pb2.SinkResponse.Result]:
    """
    Given a list of Response objects, build the corresponding list of SinkResponse.Result objects.

    Parameters:
    responses (Responses): A list of Response objects containing the results of sink operations.

    Returns:
    list[sink_pb2.SinkResponse.Result]: A list of SinkResponse.Result objects.
    """
    return [build_sink_response(rspn) for rspn in responses]


def build_sink_response(rspn: Response) -> sink_pb2.SinkResponse.Result:
    """
    Build a SinkResponse.Result object from a Response object.

    Parameters:
    rspn (Response): A Response object containing the result information of a single sink operation.

    Returns:
    sink_pb2.SinkResponse.Result: A SinkResponse.Result
     object populated with the status and id of the response.
    """
    rid = rspn.id
    if rspn.success:
        return sink_pb2.SinkResponse.Result(id=rid, status=sink_pb2.Status.SUCCESS)
    elif rspn.fallback:
        return sink_pb2.SinkResponse.Result(id=rid, status=sink_pb2.Status.FALLBACK)
    else:
        return sink_pb2.SinkResponse.Result(
            id=rid, status=sink_pb2.Status.FAILURE, err_msg=rspn.err
        )


def datum_from_sink_req(d: sink_pb2.SinkRequest) -> Datum:
    """
    Convert a SinkRequest object to a Datum object.

    Parameters:
    d (sink_pb2.SinkRequest): A SinkRequest object containing the input data.

    Returns:
    Datum: A Datum object populated with the data from the input SinkRequest object.
    """
    datum = Datum(
        keys=list(d.request.keys),
        sink_msg_id=d.request.id,
        value=d.request.value,
        event_time=d.request.event_time.ToDatetime(),
        watermark=d.request.watermark.ToDatetime(),
        headers=dict(d.request.headers),
    )
    return datum


def _create_read_handshake_response() -> sink_pb2.SinkResponse:
    """
    Create a handshake response for the Sink function.

    Returns:
    sink_pb2.SinkResponse: A SinkResponse object indicating a successful handshake.
    """
    return sink_pb2.SinkResponse(
        handshake=sink_pb2.Handshake(sot=True),
    )



from abc import ABCMeta, abstractmethod
from asyncio import Task
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import TypeVar, Callable, Union, Optional
from collections.abc import AsyncIterable

from pynumaflow.shared.asynciter import NonBlockingIterator
from pynumaflow._constants import DROP

M = TypeVar("M", bound="Message")


class WindowOperation(IntEnum):
    """
    Enumerate the type of Window operation received.
    """

    OPEN = (0,)
    CLOSE = (1,)
    APPEND = (4,)


@dataclass(init=False)
class Datum:
    """
    Class to define the important information for the event.
    Args:
        keys: the keys of the event.
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
    >>> # Example usage
    >>> from pynumaflow.reducer import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> msg_headers = {"key1": "value1", "key2": "value2"}
    >>> d = Datum(
    ...       keys=["test_key"],
    ...       value=payload,
    ...       event_time=t1,
    ...       watermark=t2,
    ...       headers=msg_headers
    ...    )
    """

    __slots__ = ("_keys", "_value", "_event_time", "_watermark", "_headers")

    _keys: list[str]
    _value: bytes
    _event_time: datetime
    _watermark: datetime
    _headers: dict[str, str]

    def __init__(
        self,
        keys: list[str],
        value: bytes,
        event_time: datetime,
        watermark: datetime,
        headers: Optional[dict[str, str]] = None,
    ):
        self._keys = keys or list()
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark
        self._headers = headers or {}

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the event"""
        return self._keys

    @property
    def value(self) -> bytes:
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event."""
        return self._watermark

    @property
    def headers(self) -> dict[str, str]:
        """Returns the headers of the event."""
        return self._headers


@dataclass(init=False)
class IntervalWindow:
    """Defines the start and end of the interval window for the event."""

    __slots__ = ("_start", "_end")

    _start: datetime
    _end: datetime

    def __init__(self, start: datetime, end: datetime):
        self._start = start
        self._end = end

    @property
    def start(self):
        """Returns the start point of the interval window."""
        return self._start

    @property
    def end(self):
        """Returns the end point of the interval window."""
        return self._end


@dataclass(init=False)
class ReduceWindow:
    """
    Defines the window for a reduce operation which includes the
    interval window along with the slot.
    """

    __slots__ = ("_window", "_slot")

    _window: IntervalWindow
    _slot: str

    def __init__(self, start: datetime, end: datetime, slot: str = ""):
        self._window = IntervalWindow(start=start, end=end)
        self._slot = slot

    @property
    def start(self):
        """Returns the start point of the interval window."""
        return self._window.start

    @property
    def end(self):
        """Returns the end point of the interval window."""
        return self._window.end

    @property
    def slot(self):
        """Returns the slot from the window"""
        return self._slot

    @property
    def window(self):
        """Return the interval window"""
        return self._window


@dataclass(init=False)
class Metadata:
    """Defines the metadata for the event."""

    __slots__ = ("_interval_window",)

    _interval_window: IntervalWindow

    def __init__(self, interval_window: IntervalWindow):
        self._interval_window = interval_window

    @property
    def interval_window(self):
        """Returns the interval window for the event."""
        return self._interval_window


@dataclass
class ReduceResult:
    """Defines the object to hold the result of reduce streaming computation."""

    __slots__ = ("_future", "_iterator", "_key", "_window", "_result_queue", "_consumer_future")

    _future: Task
    _iterator: NonBlockingIterator
    _key: list[str]
    _window: ReduceWindow
    _result_queue: NonBlockingIterator
    _consumer_future: Task

    @property
    def future(self):
        """Returns the future result of computation."""
        return self._future

    @property
    def iterator(self):
        """Returns the handle to the producer queue."""
        return self._iterator

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the partition."""
        return self._key

    @property
    def window(self) -> ReduceWindow:
        """Returns the window for the current reduce request"""
        return self._window

    @property
    def result_queue(self):
        """Returns the async queue used to write the output for the tasks"""
        return self._result_queue

    @property
    def consumer_future(self):
        """Returns the async consumer task for the result queue"""
        return self._consumer_future


@dataclass
class ReduceRequest:
    """Defines the object to hold a request for the reduce operation."""

    __slots__ = ("_operation", "_windows", "_payload")

    _operation: WindowOperation
    _windows: list[ReduceWindow]
    _payload: Datum

    def __init__(self, operation: WindowOperation, windows: list[ReduceWindow], payload: Datum):
        self._operation = operation
        self._windows = windows
        self._payload = payload

    @property
    def operation(self) -> WindowOperation:
        """Returns the future result of computation."""
        return self._operation

    @property
    def windows(self) -> list[ReduceWindow]:
        """Returns the handle to the producer queue."""
        return self._windows

    @property
    def payload(self) -> Datum:
        """Returns the payload of the window."""
        return self._payload


@dataclass(init=False)
class Message:
    """
    Basic datatype for data passing to the next vertex/vertices.

    Args:
        value: data in bytes
        keys: []string keys for vertex (optional)
        tags: []string tags for conditional forwarding (optional)
    """

    __slots__ = ("_value", "_keys", "_tags")

    _value: bytes
    _keys: list[str]
    _tags: list[str]

    def __init__(
        self,
        value: bytes,
        keys: list[str] = None,
        tags: list[str] = None,
    ):
        """
        Creates a Message object to send value to a vertex.
        """
        self._keys = keys or []
        self._tags = tags or []
        self._value = value or b""
        # self._window = window or None

    # returns the Message Object which will be dropped
    @classmethod
    def to_drop(cls: type[M]) -> M:
        return cls(b"", None, [DROP])

    @property
    def value(self) -> bytes:
        return self._value

    @property
    def keys(self) -> list[str]:
        return self._keys

    @property
    def tags(self) -> list[str]:
        return self._tags


ReduceStreamAsyncCallable = Callable[
    [list[str], AsyncIterable[Datum], NonBlockingIterator, Metadata], None
]


class ReduceStreamer(metaclass=ABCMeta):
    """
    Provides an interface to write a ReduceStreamer
    which will be exposed over a gRPC server.
    """

    def __call__(self, *args, **kwargs):
        """
        Allow to call handler function directly if class instance is sent
        as the reducer_instance.
        """
        return self.handler(*args, **kwargs)

    @abstractmethod
    async def handler(
        self,
        keys: list[str],
        datums: AsyncIterable[Datum],
        output: NonBlockingIterator,
        md: Metadata,
    ):
        """
        Implement this handler function which implements the ReduceStreamCallable interface.
        """
        pass


class _ReduceStreamBuilderClass:
    """
    Class to build a ReduceStreamer class instance.
    Used Internally

    Args:
        reducer_class: the reducer class to be used for Reduce UDF
        args: the arguments to be passed to the reducer class
        kwargs: the keyword arguments to be passed to the reducer class
    """

    def __init__(self, reducer_class: type[ReduceStreamer], args: tuple, kwargs: dict):
        self._reducer_class: type[ReduceStreamer] = reducer_class
        self._args = args
        self._kwargs = kwargs

    def create(self) -> ReduceStreamer:
        """
        Create a new ReduceStreamer instance.
        """
        return self._reducer_class(*self._args, **self._kwargs)


# ReduceStreamCallable is a callable which can be used as a handler for the Reduce UDF.
ReduceStreamCallable = Union[ReduceStreamAsyncCallable, type[ReduceStreamer]]



from pynumaflow.info.types import ServerInfo, MINIMUM_NUMAFLOW_VERSION, ContainerType
from pynumaflow.sourcetransformer.servicer._servicer import SourceTransformServicer

from pynumaflow.shared.server import start_multiproc_server

from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    SOURCE_TRANSFORMER_SOCK_PATH,
    NUM_THREADS_DEFAULT,
    UDFType,
    _PROCESS_COUNT,
    SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.sourcetransformer._dtypes import SourceTransformCallable

from pynumaflow.shared import NumaflowServer


class SourceTransformMultiProcServer(NumaflowServer):
    """
    Class for a new Source Transformer Server instance.
    """

    def __init__(
        self,
        source_transform_instance: SourceTransformCallable,
        server_count: int = _PROCESS_COUNT,
        sock_path=SOURCE_TRANSFORMER_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=SOURCE_TRANSFORMER_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Source Transformer Multiproc Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            source_transform_instance: The source transformer instance to be used for
            Source Transformer UDF
            sock_path: The UNIX socket path to be used for the server
            server_count: The number of grpc server instances to be forked for multiproc
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
                            defaults to 4 and max capped at 16

        Example invocation:
            import datetime
            import logging

            from pynumaflow.sourcetransformer import Messages, Message, Datum, SourceTransformServer

            # This is a simple User Defined Function example which receives a message,
            # applies the following
            # data transformation, and returns the message.
            # If the message event time is before year 2022, drop the message
            # with event time unchanged.
            # If it's within year 2022, update the tag to "within_year_2022" and
            # update the message event time to Jan 1st 2022.
            # Otherwise, (exclusively after year 2022), update the tag to
            # "after_year_2022" and update the


            january_first_2022 = datetime.datetime.fromtimestamp(1640995200)
            january_first_2023 = datetime.datetime.fromtimestamp(1672531200)


            def my_handler(keys: list[str], datum: Datum) -> Messages:
                val = datum.value
                event_time = datum.event_time
                messages = Messages()

                if event_time < january_first_2022:
                    logging.info("Got event time:%s, it is before 2022, so dropping", event_time)
                    messages.append(Message.to_drop(event_time))
                elif event_time < january_first_2023:
                    logging.info(
                        "Got event time:%s, it is within year 2022, so
                        forwarding to within_year_2022",
                        event_time,
                    )
                    messages.append(
                        Message(value=val, event_time=january_first_2022, tags=["within_year_2022"])
                    )
                else:
                    logging.info(
                        "Got event time:%s, it is after year 2022, so forwarding to
                        after_year_2022", event_time
                    )
                    messages.append(Message(value=val, event_time=january_first_2023,
                    tags=["after_year_2022"]))

                return messages

            if __name__ == "__main__":
                grpc_server = SourceTransformMultiProcServer(source_transform_instance=my_handler
                                                            ,server_count = 2)
                grpc_server.start()
        """
        self.sock_path = f"unix://{sock_path}"
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.max_message_size = max_message_size
        self.server_info_file = server_info_file

        self.source_transform_instance = source_transform_instance

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
            ("grpc.so_reuseport", 1),
            ("grpc.so_reuseaddr", 1),
        ]
        # Set the number of processes to be spawned to the number of CPUs or
        # the value of the parameter server_count defined by the user
        # Setting the max value to 2 * CPU count
        # Used for multiproc server
        self._process_count = min(server_count, 2 * _PROCESS_COUNT)
        self.servicer = SourceTransformServicer(handler=source_transform_instance, multiproc=True)

    def start(self):
        """
        Starts the N grpc servers gRPC serves on the with
        given max threads.
        where N = The number of CPUs or the value of the parameter server_count
        defined by the user. The max value is capped to 2 * CPU count.
        """

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[
            ContainerType.Sourcetransformer
        ]
        start_multiproc_server(
            max_threads=self.max_threads,
            servicer=self.servicer,
            process_count=self._process_count,
            server_info_file=self.server_info_file,
            server_options=self._server_options,
            udf_type=UDFType.SourceTransformer,
            server_info=serv_info,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/sinker/sink.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/sinker/sink.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"pynumaflow/proto/sinker/sink.proto\x12\x07sink.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xa3\x03\n\x0bSinkRequest\x12-\n\x07request\x18\x01 \x01(\x0b\x32\x1c.sink.v1.SinkRequest.Request\x12+\n\x06status\x18\x02 \x01(\x0b\x32\x1b.sink.v1.TransmissionStatus\x12*\n\thandshake\x18\x03 \x01(\x0b\x32\x12.sink.v1.HandshakeH\x00\x88\x01\x01\x1a\xfd\x01\n\x07Request\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\n\n\x02id\x18\x05 \x01(\t\x12:\n\x07headers\x18\x06 \x03(\x0b\x32).sink.v1.SinkRequest.Request.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0c\n\n_handshake\"\x18\n\tHandshake\x12\x0b\n\x03sot\x18\x01 \x01(\x08\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\"!\n\x12TransmissionStatus\x12\x0b\n\x03\x65ot\x18\x01 \x01(\x08\"\xfc\x01\n\x0cSinkResponse\x12-\n\x07results\x18\x01 \x03(\x0b\x32\x1c.sink.v1.SinkResponse.Result\x12*\n\thandshake\x18\x02 \x01(\x0b\x32\x12.sink.v1.HandshakeH\x00\x88\x01\x01\x12\x30\n\x06status\x18\x03 \x01(\x0b\x32\x1b.sink.v1.TransmissionStatusH\x01\x88\x01\x01\x1a\x46\n\x06Result\x12\n\n\x02id\x18\x01 \x01(\t\x12\x1f\n\x06status\x18\x02 \x01(\x0e\x32\x0f.sink.v1.Status\x12\x0f\n\x07\x65rr_msg\x18\x03 \x01(\tB\x0c\n\n_handshakeB\t\n\x07_status*0\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x12\x0c\n\x08\x46\x41LLBACK\x10\x02\x32|\n\x04Sink\x12\x39\n\x06SinkFn\x12\x14.sink.v1.SinkRequest\x1a\x15.sink.v1.SinkResponse(\x01\x30\x01\x12\x39\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x16.sink.v1.ReadyResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.sinker.sink_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SINKREQUEST_REQUEST_HEADERSENTRY']._loaded_options = None
  _globals['_SINKREQUEST_REQUEST_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_STATUS']._serialized_start=879
  _globals['_STATUS']._serialized_end=927
  _globals['_SINKREQUEST']._serialized_start=110
  _globals['_SINKREQUEST']._serialized_end=529
  _globals['_SINKREQUEST_REQUEST']._serialized_start=262
  _globals['_SINKREQUEST_REQUEST']._serialized_end=515
  _globals['_SINKREQUEST_REQUEST_HEADERSENTRY']._serialized_start=469
  _globals['_SINKREQUEST_REQUEST_HEADERSENTRY']._serialized_end=515
  _globals['_HANDSHAKE']._serialized_start=531
  _globals['_HANDSHAKE']._serialized_end=555
  _globals['_READYRESPONSE']._serialized_start=557
  _globals['_READYRESPONSE']._serialized_end=587
  _globals['_TRANSMISSIONSTATUS']._serialized_start=589
  _globals['_TRANSMISSIONSTATUS']._serialized_end=622
  _globals['_SINKRESPONSE']._serialized_start=625
  _globals['_SINKRESPONSE']._serialized_end=877
  _globals['_SINKRESPONSE_RESULT']._serialized_start=782
  _globals['_SINKRESPONSE_RESULT']._serialized_end=852
  _globals['_SINK']._serialized_start=929
  _globals['_SINK']._serialized_end=1053
# @@protoc_insertion_point(module_scope)



import asyncio
from collections.abc import AsyncIterable
from datetime import datetime
from typing import Union

from google.protobuf import timestamp_pb2
from pynumaflow._constants import (
    STREAM_EOF,
    DELIMITER,
    _LOGGER,
)
from pynumaflow.accumulator._dtypes import (
    AccumulatorResult,
    Datum,
    _AccumulatorBuilderClass,
    AccumulatorAsyncCallable,
    WindowOperation,
)
from pynumaflow.proto.accumulator import accumulator_pb2
from pynumaflow.shared.asynciter import NonBlockingIterator


def build_unique_key_name(keys):
    """
    Builds a unique key name for the given keys and window.
    The key name is used to identify the Accumulator task.
    The format is: start_time:end_time:key1:key2:...
    """
    return f"{DELIMITER.join(keys)}"


def build_window_hash(window):
    """
    Builds a hash for the given window.
    The hash is used to identify the Accumulator Window
    The format is: start_time:end_time
    """
    return f"{window.start.ToMilliseconds()}:{window.end.ToMilliseconds()}"


class TaskManager:
    """
    TaskManager is responsible for managing the Accumulator tasks.
    It is created whenever a new accumulator operation is requested.
    """

    def __init__(self, handler: Union[AccumulatorAsyncCallable, _AccumulatorBuilderClass]):
        # A dictionary to store the task information
        self.tasks: dict[str, AccumulatorResult] = {}
        # Collection for storing strong references to all running tasks.
        # Event loop only keeps a weak reference, which can cause it to
        # get lost during execution.
        self.background_tasks = set()
        # Handler for the accumulator operation
        self.__accumulator_handler = handler
        # Queue to store the results of the accumulator operation
        # This queue is used to send the results to the client
        # once the accumulator operation is completed.
        # This queue is also used to send the error/exceptions to the client
        # if the accumulator operation fails.
        self.global_result_queue = NonBlockingIterator()

    def get_unique_windows(self):
        """
        Returns the unique windows that are currently being processed
        """
        # Dict to store unique windows
        windows = dict()
        # Iterate over all the tasks and add the windows
        for task in self.tasks.values():
            window_hash = build_window_hash(task.window)
            window_found = windows.get(window_hash, None)
            # if window not seen yet, add to the dict
            if not window_found:
                windows[window_hash] = task.window
        return windows

    def get_tasks(self):
        """
        Returns the list of accumulator tasks that are
        currently being processed
        """
        return list(self.tasks.values())

    async def stream_send_eof(self):
        """
        Function used to indicate to all processing tasks that no
        more requests are expected by sending EOF message to
        local input streams of individual tasks.
        This is called when the input grpc stream is closed.
        """
        # Create a copy of the keys to avoid dictionary size change during iteration
        task_keys = list(self.tasks.keys())
        for unified_key in task_keys:
            await self.tasks[unified_key].iterator.put(STREAM_EOF)

    async def close_task(self, req):
        """
        Closes a running accumulator task for a given key.
        Based on the request we compute the unique key, and then
        signal the corresponding task for it to closure.
        The steps involve
        1. Send a signal to the local request queue of the task to stop reading
        2. Wait for the user function to complete
        3. Wait for all the results from the task to be written to the global result  queue
        4. Remove the task from the tracker
        """
        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys)
        curr_task = self.tasks.get(unified_key, None)

        if curr_task:
            await self.tasks[unified_key].iterator.put(STREAM_EOF)
            await curr_task.future
            await curr_task.consumer_future
            self.tasks.pop(unified_key)
        else:
            _LOGGER.critical("accumulator task not found", exc_info=True)
            err = BaseException("accumulator task not found")
            # Put the exception in the result queue
            await self.global_result_queue.put(err)

    async def create_task(self, req):
        """
        Creates a new accumulator task for the given request.
        Based on the request we compute a unique key, and then
        it creates a new task or appends the request to the existing task.
        """
        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys)
        curr_task = self.tasks.get(unified_key, None)

        # If the task does not exist, create a new task
        if not curr_task:
            niter = NonBlockingIterator()
            riter = niter.read_iterator()
            # Create a new result queue for the current task
            # We create a new result queue for each task, so that
            # the results of the accumulator operation can be sent to the
            # the global result queue, which in turn sends the results
            # to the client.
            res_queue = NonBlockingIterator()

            # Create a new write_to_global_queue task for the current, this will read from the
            # result queue specifically for the current task and update the
            # global result queue
            consumer = asyncio.create_task(
                self.write_to_global_queue(res_queue, self.global_result_queue, unified_key)
            )
            # Save a reference to the result of this function, to avoid a
            # task disappearing mid-execution.
            self.background_tasks.add(consumer)
            consumer.add_done_callback(self.clean_background)

            # Create a new task for the accumulator operation, this will invoke the
            # Accumulator handler with the given keys, request iterator, and window.
            task = asyncio.create_task(self.__invoke_accumulator(riter, res_queue))
            # Save a reference to the result of this function, to avoid a
            # task disappearing mid-execution.
            self.background_tasks.add(task)
            task.add_done_callback(self.clean_background)

            # Create a new AccumulatorResult object to store the task information
            curr_task = AccumulatorResult(
                task, niter, keys, res_queue, consumer, datetime.fromtimestamp(-1)
            )

            # Save the result of the accumulator operation to the task list
            self.tasks[unified_key] = curr_task

        # Put the request in the iterator
        await curr_task.iterator.put(d)

    async def send_datum_to_task(self, req):
        """
        Appends the request to the existing window reduce task.
        If the task does not exist, create it.
        """
        d = req.payload
        keys = d.keys
        unified_key = build_unique_key_name(keys)
        result = self.tasks.get(unified_key, None)
        if not result:
            await self.create_task(req)
        else:
            await result.iterator.put(d)

    async def __invoke_accumulator(
        self,
        request_iterator: AsyncIterable[Datum],
        output: NonBlockingIterator,
    ):
        """
        Invokes the UDF accumulator handler with the given keys,
        request iterator, and window. Returns the result of the
        accumulator operation.
        """
        new_instance = self.__accumulator_handler

        # If the accumulator handler is a class instance, create a new instance of it.
        # It is required for a new key to be processed by a
        # new instance of the accumulator for a given window
        # Otherwise the function handler can be called directly
        if isinstance(self.__accumulator_handler, _AccumulatorBuilderClass):
            new_instance = self.__accumulator_handler.create()
        try:
            _ = await new_instance(request_iterator, output)
            # send EOF to the output stream
            await output.put(STREAM_EOF)
        # If there is an error in the accumulator operation, log and
        # then send the error to the result queue
        except BaseException as err:
            _LOGGER.critical("panic inside accumulator handle", exc_info=True)
            # Put the exception in the result queue
            await self.global_result_queue.put(err)

    async def process_input_stream(
        self, request_iterator: AsyncIterable[accumulator_pb2.AccumulatorRequest]
    ):
        # Start iterating through the request iterator and create tasks
        # based on the operation type received.
        try:
            request_count = 0
            async for request in request_iterator:
                request_count += 1
                # check whether the request is an open or append operation
                if request.operation is int(WindowOperation.OPEN):
                    # create a new task for the open operation and
                    # put the request in the task iterator
                    await self.create_task(request)
                elif request.operation is int(WindowOperation.APPEND):
                    # append the task data to the existing task
                    # if the task does not exist, create a new task
                    await self.send_datum_to_task(request)
                elif request.operation is int(WindowOperation.CLOSE):
                    # close the current task for req
                    await self.close_task(request)
                else:
                    _LOGGER.debug(f"No operation matched for request: {request}", exc_info=True)

        # If there is an error in the accumulator operation, log and
        # then send the error to the result queue
        except BaseException as e:
            err_msg = f"Accumulator Error: {repr(e)}"
            _LOGGER.critical(err_msg, exc_info=True)
            # Put the exception in the global result queue
            await self.global_result_queue.put(e)
            return

        try:
            # send EOF to all the tasks once the request iterator is exhausted
            # This will signal the tasks to stop reading the data on their
            # respective iterators.
            await self.stream_send_eof()

            # get the list of accumulator tasks that are currently being processed
            # iterate through the tasks and wait for them to complete
            for task in self.get_tasks():
                # Once this is done, we know that the task has written all the results
                # to the local result queue
                fut = task.future
                await fut

                # Wait for the local queue to write
                # all the results of this task to the global result queue
                con_future = task.consumer_future
                await con_future
            self.tasks.clear()

            # Now send STREAM_EOF to terminate the global result queue iterator
            await self.global_result_queue.put(STREAM_EOF)
        except BaseException as e:
            err_msg = f"Accumulator Streaming Error: {repr(e)}"
            _LOGGER.critical(err_msg, exc_info=True)
            await self.global_result_queue.put(e)

    async def write_to_global_queue(
        self, input_queue: NonBlockingIterator, output_queue: NonBlockingIterator, unified_key: str
    ):
        """
        This function is used to route the messages from the
        local result queue for a given task to the global result queue.
        Once all messages are routed, it sends the window EOF messages for the same.
        """
        reader = input_queue.read_iterator()
        task = self.tasks[unified_key]

        wm: datetime = task.latest_watermark
        async for msg in reader:
            # Convert the window to a datetime object
            # Only update watermark if msg.watermark is not None
            if msg.watermark is not None and wm < msg.watermark:
                task.update_watermark(msg.watermark)
                self.tasks[unified_key] = task
                wm = msg.watermark

            # Convert datetime to protobuf timestamp
            event_time_pb = timestamp_pb2.Timestamp()
            if msg.event_time is not None:
                event_time_pb.FromDatetime(msg.event_time)

            watermark_pb = timestamp_pb2.Timestamp()
            if msg.watermark is not None:
                watermark_pb.FromDatetime(msg.watermark)

            start_dt_pb = timestamp_pb2.Timestamp()
            start_dt_pb.FromDatetime(datetime.fromtimestamp(0))

            end_dt_pb = timestamp_pb2.Timestamp()
            end_dt_pb.FromDatetime(wm)

            res = accumulator_pb2.AccumulatorResponse(
                payload=accumulator_pb2.Payload(
                    keys=msg.keys,
                    value=msg.value,
                    event_time=event_time_pb,
                    watermark=watermark_pb,
                    headers=msg.headers,
                    id=msg.id,
                ),
                window=accumulator_pb2.KeyedWindow(
                    start=start_dt_pb, end=end_dt_pb, slot="slot-0", keys=task.keys
                ),
                EOF=False,
                tags=msg.tags,
            )
            await output_queue.put(res)
        # send EOF
        start_eof_pb = timestamp_pb2.Timestamp()
        start_eof_pb.FromDatetime(datetime.fromtimestamp(0))

        end_eof_pb = timestamp_pb2.Timestamp()
        end_eof_pb.FromDatetime(wm)

        res = accumulator_pb2.AccumulatorResponse(
            window=accumulator_pb2.KeyedWindow(
                start=start_eof_pb, end=end_eof_pb, slot="slot-0", keys=task.keys
            ),
            EOF=True,
        )
        await output_queue.put(res)

    def clean_background(self, task):
        """
        Remove the task from the background tasks collection
        """
        self.background_tasks.remove(task)



from abc import abstractmethod, ABCMeta
from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar, Optional, Callable, Union
from collections.abc import AsyncIterable, Awaitable
from collections.abc import Sequence, Iterator
from warnings import warn

R = TypeVar("R", bound="Response")
Rs = TypeVar("Rs", bound="Responses")


@dataclass
class Response:
    """
    Basic datatype for UDSink response.

    Args:
        id: the id of the event.
        success: boolean indicating whether the event was successfully processed.
        err: error message if the event was not successfully processed.
        fallback: fallback is true if the message to be sent to the fallback sink.
    """

    id: str
    success: bool
    err: Optional[str]
    fallback: bool

    __slots__ = ("id", "success", "err", "fallback")

    # as_success creates a successful Response with the given id.
    # The Success field is set to true.
    @classmethod
    def as_success(cls: type[R], id_: str) -> R:
        return Response(id=id_, success=True, err=None, fallback=False)

    # as_failure creates a failed Response with the given id and error message.
    # The success field is set to false and the err field is set to the provided error message.
    @classmethod
    def as_failure(cls: type[R], id_: str, err_msg: str) -> R:
        return Response(id=id_, success=False, err=err_msg, fallback=False)

    # as_fallback creates a Response with the fallback field set to true.
    # This indicates that the message should be sent to the fallback sink.
    @classmethod
    def as_fallback(cls: type[R], id_: str) -> R:
        return Response(id=id_, fallback=True, err=None, success=False)


class Responses(Sequence[R]):
    """
    Container to hold a list of Response instances.

    Args:
        responses: list of Response instances.
    """

    __slots__ = ("_responses",)

    def __init__(self, *responses: R):
        self._responses = list(responses) or []

    def __str__(self) -> str:
        return str(self._responses)

    def __repr__(self) -> str:
        return str(self)

    def __len__(self) -> int:
        return len(self._responses)

    def __iter__(self) -> Iterator[R]:
        return iter(self._responses)

    def __getitem__(self, index: int) -> R:
        if isinstance(index, slice):
            raise TypeError("Slicing is not supported for Responses")
        return self._responses[index]

    def append(self, response: R) -> None:
        self._responses.append(response)

    def items(self) -> list[R]:
        warn(
            "Using items is deprecated and will be removed in v0.5. "
            "Iterate or index the Responses object instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._responses


@dataclass(init=False, repr=False)
class Datum:
    """
    Class to define the important information for the event.
    Args:
        keys: the keys of the event.
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
        headers: the headers of the event.
    >>> # Example usage
    >>> from pynumaflow.sinker import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> msg_headers = {"key1": "value1", "key2": "value2"}
    >>> msg_id = "test_id"
    >>> output_keys = ["test_key"]
    >>> d = Datum(
    ...       keys=output_keys,
    ...       sink_msg_id=msg_id,
    ...       value=payload,
    ...       event_time=t1,
    ...       watermark=t2,
    ...       headers=msg_headers
    ...    )
    """

    __slots__ = ("_keys", "_id", "_value", "_event_time", "_watermark", "_headers")

    _keys: list[str]
    _id: str
    _value: bytes
    _event_time: datetime
    _watermark: datetime
    _headers: dict[str, str]

    def __init__(
        self,
        keys: list[str],
        sink_msg_id: str,
        value: bytes,
        event_time: datetime,
        watermark: datetime,
        headers: Optional[dict[str, str]] = None,
    ):
        self._keys = keys
        self._id = sink_msg_id or ""
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark
        self._headers = headers or {}

    def __str__(self):
        value_string = self._value.decode("utf-8")
        return (
            f"keys: {self._keys}, "
            f"id: {self._id}, value: {value_string}, "
            f"event_time: {str(self._event_time)}, "
            f"watermark: {str(self._watermark)}, "
            f"headers: {self._headers}"
        )

    def __repr__(self):
        return str(self)

    @property
    def id(self) -> str:
        """Returns the id of the event."""
        return self._id

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the event."""
        return self._keys

    @property
    def value(self) -> bytes:
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event."""
        return self._watermark

    @property
    def headers(self) -> dict[str, str]:
        """Returns the headers of the event."""
        return self._headers.copy()


class Sinker(metaclass=ABCMeta):
    """
    Provides an interface to write a Sinker
    which will be exposed over a gRPC server.

    """

    def __call__(self, *args, **kwargs):
        """
        Allow to call handler function directly if class instance is sent
        as the sinker_instance.
        """
        return self.handler(*args, **kwargs)

    @abstractmethod
    def handler(self, datums: Iterator[Datum]) -> Responses:
        """
        Implement this handler function which implements the SinkCallable interface.
        """
        pass


# SinkSyncCallable is a callable which can be used as a handler for the Synchronous UDSink.
SinkHandlerCallable = Callable[[Iterator[Datum]], Responses]
SinkSyncCallable = Union[Sinker, SinkHandlerCallable]

# SinkAsyncCallable is a callable which can be used as a handler for the Asynchronous UDSink.
AsyncSinkHandlerCallable = Callable[[AsyncIterable[Datum]], Awaitable[Responses]]
SinkAsyncCallable = Union[Sinker, AsyncSinkHandlerCallable]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/reducer/reduce.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/reducer/reduce.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%pynumaflow/proto/reducer/reduce.proto\x12\treduce.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\x98\x04\n\rReduceRequest\x12\x31\n\x07payload\x18\x01 \x01(\x0b\x32 .reduce.v1.ReduceRequest.Payload\x12;\n\toperation\x18\x02 \x01(\x0b\x32(.reduce.v1.ReduceRequest.WindowOperation\x1a\x9e\x01\n\x0fWindowOperation\x12=\n\x05\x65vent\x18\x01 \x01(\x0e\x32..reduce.v1.ReduceRequest.WindowOperation.Event\x12\"\n\x07windows\x18\x02 \x03(\x0b\x32\x11.reduce.v1.Window\"(\n\x05\x45vent\x12\x08\n\x04OPEN\x10\x00\x12\t\n\x05\x43LOSE\x10\x01\x12\n\n\x06\x41PPEND\x10\x04\x1a\xf5\x01\n\x07Payload\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12>\n\x07headers\x18\x05 \x03(\x0b\x32-.reduce.v1.ReduceRequest.Payload.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"j\n\x06Window\x12)\n\x05start\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03\x65nd\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04slot\x18\x03 \x01(\t\"\xa7\x01\n\x0eReduceResponse\x12\x30\n\x06result\x18\x01 \x01(\x0b\x32 .reduce.v1.ReduceResponse.Result\x12!\n\x06window\x18\x02 \x01(\x0b\x32\x11.reduce.v1.Window\x12\x0b\n\x03\x45OF\x18\x03 \x01(\x08\x1a\x33\n\x06Result\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x0c\n\x04tags\x18\x03 \x03(\t\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\x8a\x01\n\x06Reduce\x12\x43\n\x08ReduceFn\x12\x18.reduce.v1.ReduceRequest\x1a\x19.reduce.v1.ReduceResponse(\x01\x30\x01\x12;\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x18.reduce.v1.ReadyResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.reducer.reduce_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_REDUCEREQUEST_PAYLOAD_HEADERSENTRY']._loaded_options = None
  _globals['_REDUCEREQUEST_PAYLOAD_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_REDUCEREQUEST']._serialized_start=115
  _globals['_REDUCEREQUEST']._serialized_end=651
  _globals['_REDUCEREQUEST_WINDOWOPERATION']._serialized_start=245
  _globals['_REDUCEREQUEST_WINDOWOPERATION']._serialized_end=403
  _globals['_REDUCEREQUEST_WINDOWOPERATION_EVENT']._serialized_start=363
  _globals['_REDUCEREQUEST_WINDOWOPERATION_EVENT']._serialized_end=403
  _globals['_REDUCEREQUEST_PAYLOAD']._serialized_start=406
  _globals['_REDUCEREQUEST_PAYLOAD']._serialized_end=651
  _globals['_REDUCEREQUEST_PAYLOAD_HEADERSENTRY']._serialized_start=605
  _globals['_REDUCEREQUEST_PAYLOAD_HEADERSENTRY']._serialized_end=651
  _globals['_WINDOW']._serialized_start=653
  _globals['_WINDOW']._serialized_end=759
  _globals['_REDUCERESPONSE']._serialized_start=762
  _globals['_REDUCERESPONSE']._serialized_end=929
  _globals['_REDUCERESPONSE_RESULT']._serialized_start=878
  _globals['_REDUCERESPONSE_RESULT']._serialized_end=929
  _globals['_READYRESPONSE']._serialized_start=931
  _globals['_READYRESPONSE']._serialized_end=961
  _globals['_REDUCE']._serialized_start=964
  _globals['_REDUCE']._serialized_end=1102
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: pynumaflow/proto/accumulator/accumulator.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'pynumaflow/proto/accumulator/accumulator.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.pynumaflow/proto/accumulator/accumulator.proto\x12\x0e\x61\x63\x63umulator.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xf8\x01\n\x07Payload\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\n\n\x02id\x18\x05 \x01(\t\x12\x35\n\x07headers\x18\x06 \x03(\x0b\x32$.accumulator.v1.Payload.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xbe\x02\n\x12\x41\x63\x63umulatorRequest\x12(\n\x07payload\x18\x01 \x01(\x0b\x32\x17.accumulator.v1.Payload\x12\x45\n\toperation\x18\x02 \x01(\x0b\x32\x32.accumulator.v1.AccumulatorRequest.WindowOperation\x1a\xb6\x01\n\x0fWindowOperation\x12G\n\x05\x65vent\x18\x01 \x01(\x0e\x32\x38.accumulator.v1.AccumulatorRequest.WindowOperation.Event\x12\x30\n\x0bkeyedWindow\x18\x02 \x01(\x0b\x32\x1b.accumulator.v1.KeyedWindow\"(\n\x05\x45vent\x12\x08\n\x04OPEN\x10\x00\x12\t\n\x05\x43LOSE\x10\x01\x12\n\n\x06\x41PPEND\x10\x02\"}\n\x0bKeyedWindow\x12)\n\x05start\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03\x65nd\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04slot\x18\x03 \x01(\t\x12\x0c\n\x04keys\x18\x04 \x03(\t\"\x87\x01\n\x13\x41\x63\x63umulatorResponse\x12(\n\x07payload\x18\x01 \x01(\x0b\x32\x17.accumulator.v1.Payload\x12+\n\x06window\x18\x02 \x01(\x0b\x32\x1b.accumulator.v1.KeyedWindow\x12\x0c\n\x04tags\x18\x03 \x03(\t\x12\x0b\n\x03\x45OF\x18\x04 \x01(\x08\"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\xac\x01\n\x0b\x41\x63\x63umulator\x12[\n\x0c\x41\x63\x63umulateFn\x12\".accumulator.v1.AccumulatorRequest\x1a#.accumulator.v1.AccumulatorResponse(\x01\x30\x01\x12@\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x1d.accumulator.v1.ReadyResponseBd\n#io.numaproj.numaflow.accumulator.v1Z=github.com/numaproj/numaflow-go/pkg/apis/proto/accumulator/v1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pynumaflow.proto.accumulator.accumulator_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n#io.numaproj.numaflow.accumulator.v1Z=github.com/numaproj/numaflow-go/pkg/apis/proto/accumulator/v1'
  _globals['_PAYLOAD_HEADERSENTRY']._loaded_options = None
  _globals['_PAYLOAD_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_PAYLOAD']._serialized_start=129
  _globals['_PAYLOAD']._serialized_end=377
  _globals['_PAYLOAD_HEADERSENTRY']._serialized_start=331
  _globals['_PAYLOAD_HEADERSENTRY']._serialized_end=377
  _globals['_ACCUMULATORREQUEST']._serialized_start=380
  _globals['_ACCUMULATORREQUEST']._serialized_end=698
  _globals['_ACCUMULATORREQUEST_WINDOWOPERATION']._serialized_start=516
  _globals['_ACCUMULATORREQUEST_WINDOWOPERATION']._serialized_end=698
  _globals['_ACCUMULATORREQUEST_WINDOWOPERATION_EVENT']._serialized_start=658
  _globals['_ACCUMULATORREQUEST_WINDOWOPERATION_EVENT']._serialized_end=698
  _globals['_KEYEDWINDOW']._serialized_start=700
  _globals['_KEYEDWINDOW']._serialized_end=825
  _globals['_ACCUMULATORRESPONSE']._serialized_start=828
  _globals['_ACCUMULATORRESPONSE']._serialized_end=963
  _globals['_READYRESPONSE']._serialized_start=965
  _globals['_READYRESPONSE']._serialized_end=995
  _globals['_ACCUMULATOR']._serialized_start=998
  _globals['_ACCUMULATOR']._serialized_end=1170
# @@protoc_insertion_point(module_scope)



import inspect
from typing import Optional

import aiorun
import grpc

from pynumaflow.info.types import ServerInfo, MINIMUM_NUMAFLOW_VERSION, ContainerType
from pynumaflow.proto.reducer import reduce_pb2_grpc

from pynumaflow.reducer.servicer.async_servicer import AsyncReduceServicer

from pynumaflow._constants import (
    REDUCE_SOCK_PATH,
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    REDUCE_SERVER_INFO_FILE_PATH,
    MAX_NUM_THREADS,
)

from pynumaflow.reducer._dtypes import (
    ReduceCallable,
    _ReduceBuilderClass,
    Reducer,
)

from pynumaflow.shared.server import NumaflowServer, check_instance, start_async_server


def get_handler(
    reducer_handler: ReduceCallable, init_args: tuple = (), init_kwargs: Optional[dict] = None
):
    """
    Get the correct handler type based on the arguments passed
    """
    if inspect.isfunction(reducer_handler):
        if len(init_args) > 0 or len(init_kwargs) > 0:
            # if the init_args or init_kwargs are passed, then the reducer_instance
            # can only be of class Reducer type
            raise TypeError("Cannot pass function handler with init args or kwargs")
        # return the function handler
        return reducer_handler
    elif not check_instance(reducer_handler, Reducer) and issubclass(reducer_handler, Reducer):
        # if handler is type of Class Reducer, create a new instance of
        # a ReducerBuilderClass
        return _ReduceBuilderClass(reducer_handler, init_args, init_kwargs)
    else:
        _LOGGER.error(
            _error_msg := f"Invalid Class Type {reducer_handler}: "
            f"Please make sure the class type is passed, and it is a subclass of Reducer"
        )
        raise TypeError(_error_msg)


class ReduceAsyncServer(NumaflowServer):
    """
    Class for a new Reduce Server instance.
    A new servicer instance is created and attached to the server.
    The server instance is returned.
    Args:
        reducer_instance: The reducer instance to be used for Reduce UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16
    Example invocation:
        import os
        from collections.abc import AsyncIterable
        from pynumaflow.reducer import Messages, Message, Datum, Metadata,
        ReduceAsyncServer, Reducer

        class ReduceCounter(Reducer):
            def __init__(self, counter):
                self.counter = counter

            async def handler(
                self, keys: list[str], datums: AsyncIterable[Datum], md: Metadata
            ) -> Messages:
                interval_window = md.interval_window
                self.counter = 0
                async for _ in datums:
                    self.counter += 1
                msg = (
                    f"counter:{self.counter} interval_window_start:{interval_window.start} "
                    f"interval_window_end:{interval_window.end}"
                )
                return Messages(Message(str.encode(msg), keys=keys))

        async def reduce_handler(keys: list[str],
                                datums: AsyncIterable[Datum],
                                md: Metadata) -> Messages:
            interval_window = md.interval_window
            counter = 0
            async for _ in datums:
                counter += 1
            msg = (
                f"counter:{counter} interval_window_start:{interval_window.start} "
                f"interval_window_end:{interval_window.end}"
            )
            return Messages(Message(str.encode(msg), keys=keys))

        if __name__ == "__main__":
            invoke = os.getenv("INVOKE", "func_handler")
            if invoke == "class":
                # Here we are using the class instance as the reducer_instance
                # which will be used to invoke the handler function.
                # We are passing the init_args for the class instance.
                grpc_server = ReduceAsyncServer(ReduceCounter, init_args=(0,))
            else:
                # Here we are using the handler function directly as the reducer_instance.
                grpc_server = ReduceAsyncServer(reduce_handler)
            grpc_server.start()

    """

    def __init__(
        self,
        reducer_instance: ReduceCallable,
        init_args: tuple = (),
        init_kwargs: dict = None,
        sock_path=REDUCE_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=REDUCE_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Reduce Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
        reducer_instance: The reducer instance to be used for Reduce UDF
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                        defaults to 4 and max capped at 16
        server_type: The type of server to be used
        """
        if init_kwargs is None:
            init_kwargs = {}
        self.reducer_handler = get_handler(reducer_instance, init_args, init_kwargs)
        self.sock_path = f"unix://{sock_path}"
        self.max_message_size = max_message_size
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.server_info_file = server_info_file

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        # Get the servicer instance for the async server
        self.servicer = AsyncReduceServicer(self.reducer_handler)

    def start(self):
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        _LOGGER.info(
            "Starting Async Reduce Server",
        )
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new async server instance and add the servicer to it
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        reduce_servicer = self.servicer
        reduce_pb2_grpc.add_ReduceServicer_to_server(reduce_servicer, server)

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Reducer]
        # Start the async server
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



import inspect
from typing import Optional

import aiorun
import grpc

from pynumaflow.accumulator.servicer.async_servicer import AsyncAccumulatorServicer
from pynumaflow.info.types import ServerInfo, ContainerType, MINIMUM_NUMAFLOW_VERSION
from pynumaflow.proto.accumulator import accumulator_pb2_grpc


from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    NUM_THREADS_DEFAULT,
    _LOGGER,
    MAX_NUM_THREADS,
    ACCUMULATOR_SOCK_PATH,
    ACCUMULATOR_SERVER_INFO_FILE_PATH,
)

from pynumaflow.accumulator._dtypes import (
    AccumulatorStreamCallable,
    _AccumulatorBuilderClass,
    Accumulator,
)

from pynumaflow.shared.server import NumaflowServer, check_instance, start_async_server


def get_handler(
    accumulator_handler: AccumulatorStreamCallable,
    init_args: tuple = (),
    init_kwargs: Optional[dict] = None,
):
    """
    Get the correct handler type based on the arguments passed
    """
    if inspect.isfunction(accumulator_handler):
        if init_args or init_kwargs:
            # if the init_args or init_kwargs are passed, then the accumulator_instance
            # can only be of class Accumulator type
            raise TypeError("Cannot pass function handler with init args or kwargs")
        # return the function handler
        return accumulator_handler
    elif not check_instance(accumulator_handler, Accumulator) and issubclass(
        accumulator_handler, Accumulator
    ):
        # if handler is type of Class Accumulator, create a new instance of
        # a AccumulatorBuilderClass
        return _AccumulatorBuilderClass(accumulator_handler, init_args, init_kwargs)
    else:
        _LOGGER.error(
            _error_msg := f"Invalid Class Type {accumulator_handler}: "
            f"Please make sure the class type is passed, and it is a subclass of Accumulator"
        )
        raise TypeError(_error_msg)


class AccumulatorAsyncServer(NumaflowServer):
    """
    Class for a new Accumulator Server instance.
    A new servicer instance is created and attached to the server.
    The server instance is returned.
    Args:
        accumulator_instance: The accumulator instance to be used for
                Accumulator UDF
        init_args: The arguments to be passed to the accumulator_handler
        init_kwargs: The keyword arguments to be passed to the
            accumulator_handler
        sock_path: The UNIX socket path to be used for the server
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
        defaults to 4 and max capped at 16
        server_info_file: The path to the server info file
    Example invocation:
        import os
        from collections.abc import AsyncIterable
        from datetime import datetime

        from pynumaflow.accumulator import Accumulator, AccumulatorAsyncServer
        from pynumaflow.accumulator import (
            Message,
            Datum,
        )
        from pynumaflow.shared.asynciter import NonBlockingIterator

        class StreamSorter(Accumulator):
            def __init__(self, counter):
                self.latest_wm = datetime.fromtimestamp(-1)
                self.sorted_buffer: list[Datum] = []

            async def handler(
                self,
                datums: AsyncIterable[Datum],
                output: NonBlockingIterator,
            ):
                async for _ in datums:
                    # Process the datums and send output
                    if datum.watermark and datum.watermark > self.latest_wm:
                        self.latest_wm = datum.watermark
                        await self.flush_buffer(output)

                    self.insert_sorted(datum)

            def insert_sorted(self, datum: Datum):
                # Binary insert to keep sorted buffer in order
                left, right = 0, len(self.sorted_buffer)
                while left < right:
                    mid = (left + right) // 2
                    if self.sorted_buffer[mid].event_time > datum.event_time:
                        right = mid
                    else:
                        left = mid + 1
                self.sorted_buffer.insert(left, datum)

            async def flush_buffer(self, output: NonBlockingIterator):
                i = 0
                for datum in self.sorted_buffer:
                    if datum.event_time > self.latest_wm:
                        break
                    await output.put(Message.from_datum(datum))
                    i += 1
                # Remove flushed items
                self.sorted_buffer = self.sorted_buffer[i:]


        if __name__ == "__main__":
            grpc_server = AccumulatorAsyncServer(StreamSorter)
            grpc_server.start()

    """

    def __init__(
        self,
        accumulator_instance: AccumulatorStreamCallable,
        init_args: tuple = (),
        init_kwargs: dict = None,
        sock_path=ACCUMULATOR_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=NUM_THREADS_DEFAULT,
        server_info_file=ACCUMULATOR_SERVER_INFO_FILE_PATH,
    ):
        """
        Create a new grpc Accumulator Server instance.
        A new servicer instance is created and attached to the server.
        The server instance is returned.
        Args:
            accumulator_instance: The Accumulator instance to be used for
                    Accumulator UDF
            init_args: The arguments to be passed to the accumulator_handler
            init_kwargs: The keyword arguments to be passed to the
                accumulator_handler
            sock_path: The UNIX socket path to be used for the server
            max_message_size: The max message size in bytes the server can receive and send
            max_threads: The max number of threads to be spawned;
            defaults to 4 and max capped at 16
            server_info_file: The path to the server info file
        """
        if init_kwargs is None:
            init_kwargs = {}
        self.accumulator_handler = get_handler(accumulator_instance, init_args, init_kwargs)
        self.sock_path = f"unix://{sock_path}"
        self.max_message_size = max_message_size
        self.max_threads = min(max_threads, MAX_NUM_THREADS)
        self.server_info_file = server_info_file

        self._server_options = [
            ("grpc.max_send_message_length", self.max_message_size),
            ("grpc.max_receive_message_length", self.max_message_size),
        ]
        # Get the servicer instance for the async server
        self.servicer = AsyncAccumulatorServicer(self.accumulator_handler)

    def start(self):
        """
        Starter function for the Async server class, need a separate caller
        so that all the async coroutines can be started from a single context
        """
        _LOGGER.info(
            "Starting Async Accumulator Server",
        )
        aiorun.run(self.aexec(), use_uvloop=True)

    async def aexec(self):
        """
        Starts the Async gRPC server on the given UNIX socket with
        given max threads.
        """
        # As the server is async, we need to create a new server instance in the
        # same thread as the event loop so that all the async calls are made in the
        # same context
        # Create a new async server instance and add the servicer to it
        server = grpc.aio.server(options=self._server_options)
        server.add_insecure_port(self.sock_path)
        accumulator_pb2_grpc.add_AccumulatorServicer_to_server(self.servicer, server)

        serv_info = ServerInfo.get_default_server_info()
        serv_info.minimum_numaflow_version = MINIMUM_NUMAFLOW_VERSION[ContainerType.Accumulator]
        await start_async_server(
            server_async=server,
            sock_path=self.sock_path,
            max_threads=self.max_threads,
            cleanup_coroutines=list(),
            server_info_file=self.server_info_file,
            server_info=serv_info,
        )



from pynumaflow._constants import DROP

from pynumaflow.mapstreamer._dtypes import Message, Messages, Datum, MapStreamer
from pynumaflow.mapstreamer.async_server import MapStreamAsyncServer

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "MapStreamAsyncServer",
    "MapStreamer",
]
